# Copyright 2023 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
import itertools
import logging
import respx
import pytest

from planet import exceptions, subscription_request, specs

LOGGER = logging.getLogger(__name__)


def test_build_request_success(geom_geojson):
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    delivery = {
        "type": "amazon_s3",
        "parameters": {
            "aws_access_key_id": "keyid",
            "aws_secret_access_key": "accesskey",
            "bucket": "bucket",
            "aws_region": "region"
        }
    }

    tool = {"type": "clip", "parameters": {"aoi": geom_geojson}}

    notifications = {"webhook": {"url": "url", "topics": ['delivery.success']}}

    res = subscription_request.build_request('test',
                                             source=source,
                                             delivery=delivery,
                                             tools=[tool],
                                             notifications=notifications)

    expected = {
        "name": "test",
        "source": source,
        "delivery": delivery,
        "tools": [tool],
        "notifications": notifications
    }

    assert res == expected


def test_build_request_clip_to_source_success(geom_geojson):
    """Without a clip tool we can clip to source."""
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }
    req = subscription_request.build_request(
        'test',
        source=source,
        delivery={},
        tools=[{
            'type': 'hammer'
        }],
        clip_to_source=True,
    )
    assert req["tools"][1]["type"] == "clip"
    assert req["tools"][1]["parameters"]["aoi"] == geom_geojson


def test_build_request_clip_to_source_failure(geom_geojson):
    """With a clip tool we can not clip to source."""
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }
    with pytest.raises(exceptions.ClientError):
        subscription_request.build_request(
            'test',
            source=source,
            delivery={},
            tools=[{
                'type': 'clip'
            }, {
                'type': 'hammer'
            }],
            clip_to_source=True,
        )


def test_build_request_host_sentinel_hub_no_collection(geom_geojson):
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    hosting = {"type": "sentinel-hub"}

    res = subscription_request.build_request('test',
                                             source=source,
                                             hosting=hosting)

    expected = {"name": "test", "source": source, "hosting": hosting}

    assert res == expected


def test_build_request_host_sentinel_hub_with_collection(geom_geojson):
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    hosting = {
        "type": "sentinel-hub",
        "parameters": {
            "collection_id": "4c9af036-4274-4a97-bf0d-eb2a7853330d"
        }
    }

    res = subscription_request.build_request('test',
                                             source=source,
                                             hosting=hosting)

    expected = {"name": "test", "source": source, "hosting": hosting}

    assert res == expected


def test_build_request_host_sentinel_hub_create_configuration(geom_geojson):
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    hosting = {
        "type": "sentinel-hub", "parameters": {
            "create_configuration": True
        }
    }

    res = subscription_request.build_request('test',
                                             source=source,
                                             hosting=hosting)

    expected = {"name": "test", "source": source, "hosting": hosting}

    assert res == expected


def test_build_request_host_sentinel_hub_collection_configuration(
        geom_geojson):
    source = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    hosting = {
        "type": "sentinel-hub",
        "parameters": {
            "collection_id": "1234", "create_configuration": True
        }
    }

    res = subscription_request.build_request('test',
                                             source=source,
                                             hosting=hosting)

    expected = {"name": "test", "source": source, "hosting": hosting}

    assert res == expected


@respx.mock
def test_catalog_source_success(geom_geojson, mock_bundles):
    res = subscription_request.catalog_source(
        item_types=["PSScene"],
        asset_types=["ortho_analytic_4b"],
        geometry=geom_geojson,
        start_time=datetime(2021, 3, 1),
        end_time=datetime(2023, 11, 1),
        rrule="FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
    )

    expected = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "end_time": "2023-11-01T00:00:00Z",
            "rrule": "FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    assert res == expected


@respx.mock
def test_catalog_source_featurecollection(featurecollection_geojson,
                                          geom_geojson,
                                          mock_bundles):
    """geojson specified as featurecollection is simplified down to just
    the geometry"""
    res = subscription_request.catalog_source(
        item_types=["PSScene"],
        asset_types=["ortho_analytic_4b"],
        geometry=featurecollection_geojson,
        start_time=datetime(2021, 3, 1),
    )

    expected = {
        "parameters": {
            "geometry": geom_geojson,
            "start_time": "2021-03-01T00:00:00Z",
            "item_types": ["PSScene"],
            "asset_types": ["ortho_analytic_4b"]
        }
    }

    assert res == expected


@respx.mock
def test_catalog_source_invalid_start_time(geom_geojson, mock_bundles):
    with pytest.raises(exceptions.ClientError):
        subscription_request.catalog_source(
            item_types=["PSScene"],
            asset_types=["ortho_analytic_4b"],
            geometry=geom_geojson,
            start_time='invalid',
            end_time=datetime(2023, 11, 1),
            rrule="FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
        )


def test_amazon_s3_success():
    res = subscription_request.amazon_s3(aws_access_key_id='keyid',
                                         aws_secret_access_key='accesskey',
                                         bucket='bucket',
                                         aws_region='region')

    assert res == {
        "type": "amazon_s3",
        "parameters": {
            "aws_access_key_id": "keyid",
            "aws_secret_access_key": "accesskey",
            "bucket": "bucket",
            "aws_region": "region"
        }
    }


def test_amazon_s3_path_prefix_success():
    res = subscription_request.amazon_s3(aws_access_key_id='keyid',
                                         aws_secret_access_key='accesskey',
                                         bucket='bucket',
                                         aws_region='region',
                                         path_prefix="prefix")

    assert res == {
        "type": "amazon_s3",
        "parameters": {
            "aws_access_key_id": "keyid",
            "aws_secret_access_key": "accesskey",
            "bucket": "bucket",
            "aws_region": "region",
            "path_prefix": "prefix"
        }
    }


def test_azure_blob_storage_success():
    res = subscription_request.azure_blob_storage(
        account='act',
        container='container',
        sas_token='sastoken',
        storage_endpoint_suffix='suffix')

    assert res == {
        "type": "azure_blob_storage",
        "parameters": {
            "account": "act",
            "container": "container",
            "sas_token": "sastoken",
            "storage_endpoint_suffix": "suffix"
        }
    }


def test_azure_blob_storage_path_prefix_success():
    res = subscription_request.azure_blob_storage(
        account='act',
        container='container',
        sas_token='sastoken',
        storage_endpoint_suffix='suffix',
        path_prefix="prefix")

    assert res == {
        "type": "azure_blob_storage",
        "parameters": {
            "account": "act",
            "container": "container",
            "sas_token": "sastoken",
            "storage_endpoint_suffix": "suffix",
            "path_prefix": "prefix"
        }
    }


def test_google_earth_engine_success():
    res = subscription_request.google_earth_engine(project='project',
                                                   collection='collection',
                                                   credentials='cred')

    assert res == {
        "type": "google_earth_engine",
        "parameters": {
            "project": "project",
            "collection": "collection",
            "credentials": "cred"
        }
    }


def test_google_cloud_storage_success():
    res = subscription_request.google_cloud_storage(credentials='cred',
                                                    bucket='bucket')

    assert res == {
        "type": "google_cloud_storage",
        "parameters": {
            "bucket": "bucket", "credentials": "cred"
        }
    }


def test_google_cloud_storage_path_prefix_success():
    res = subscription_request.google_cloud_storage(credentials='cred',
                                                    bucket='bucket',
                                                    path_prefix="prefix")

    assert res == {
        "type": "google_cloud_storage",
        "parameters": {
            "bucket": "bucket", "credentials": "cred", "path_prefix": "prefix"
        }
    }


def test_oracle_cloud_storage_success():
    res = subscription_request.oracle_cloud_storage(
        customer_access_key_id='keyid',
        customer_secret_key='secretkey',
        bucket='bucket',
        region='region',
        namespace='namespace')

    assert res == {
        "type": "oracle_cloud_storage",
        "parameters": {
            "customer_access_key_id": "keyid",
            "customer_secret_key": "secretkey",
            "bucket": "bucket",
            "region": "region",
            "namespace": "namespace"
        }
    }


def test_oracle_cloud_storage_path_prefix_success():
    res = subscription_request.oracle_cloud_storage(
        customer_access_key_id='keyid',
        customer_secret_key='secretkey',
        bucket='bucket',
        region='region',
        namespace='namespace',
        path_prefix="prefix")

    assert res == {
        "type": "oracle_cloud_storage",
        "parameters": {
            "customer_access_key_id": "keyid",
            "customer_secret_key": "secretkey",
            "bucket": "bucket",
            "region": "region",
            "namespace": "namespace",
            "path_prefix": "prefix"
        }
    }


def test_notifications_success():
    topics = ['delivery.success']
    notifications_config = subscription_request.notifications(url='url',
                                                              topics=topics)

    expected = {"webhook": {"url": "url", "topics": topics}}
    assert notifications_config == expected


def test_notifications_invalid_topics():
    with pytest.raises(exceptions.ClientError):
        subscription_request.notifications(url='url', topics=['invalid'])


def test_band_math_tool_success():
    res = subscription_request.band_math_tool(b1='b1', b2='arctan(b1)')

    expected = {
        "type": "bandmath",
        "parameters": {
            "b1": "b1",
            "b2": "arctan(b1)",
            "pixel_type": specs.BAND_MATH_PIXEL_TYPE_DEFAULT
        }
    }
    assert res == expected


def test_band_math_tool_invalid_pixel_type():
    with pytest.raises(exceptions.ClientError):
        subscription_request.band_math_tool(b1='b1',
                                            b2='arctan(b1)',
                                            pixel_type="invalid")


def test_clip_tool_success(geom_geojson):
    res = subscription_request.clip_tool(geom_geojson)
    expected = {"type": "clip", "parameters": {"aoi": geom_geojson}}
    assert res == expected


def test_clip_tool_invalid_type(point_geom_geojson):
    with pytest.raises(exceptions.ClientError):
        subscription_request.clip_tool(point_geom_geojson)


def test_file_format_tool_success():
    res = subscription_request.file_format_tool('COG')

    expected = {"type": "file_format", "parameters": {"format": "COG"}}
    assert res == expected


def test_file_format_tool_invalid_format():
    with pytest.raises(exceptions.ClientError):
        subscription_request.file_format_tool('invalid')


def test_harmonize_tool_success():
    res = subscription_request.harmonize_tool('Sentinel-2')

    expected = {
        "type": "harmonize", "parameters": {
            "target_sensor": "Sentinel-2"
        }
    }
    assert res == expected


def test_harmonize_tool_invalid_target_sensor():
    with pytest.raises(exceptions.ClientError):
        subscription_request.harmonize_tool('invalid')


def test_reproject_tool_success():
    res = subscription_request.reproject_tool('EPSG:4326',
                                              kernel='near',
                                              resolution=0.5)

    expected = {
        "type": "reproject",
        "parameters": {
            "projection": "EPSG:4326", "kernel": "near", "resolution": 0.5
        }
    }
    assert res == expected


def test_reproject_tool_invalid_kernel():
    with pytest.raises(exceptions.ClientError):
        subscription_request.reproject_tool('EPSG:4326',
                                            kernel='invalid',
                                            resolution=0.5)


def test_toar_tool_success():
    res = subscription_request.toar_tool(12345)

    expected = {"type": "toar", "parameters": {"scale_factor": 12345}}
    assert res == expected


@pytest.mark.parametrize(
    "var_type, var_id",
    [
        ("biomass_proxy", "BIOMASS-PROXY_V3.0_10"),  # actual real type and id.
        ("var1", "VAR1-ABCD"),  # nonsense type and id
        (None, "BIOMASS-PROXY_V3.0_10"),  # None type with valid id
    ])
def test_pv_source_success(geom_geojson, var_type, var_id):
    """Configure a planetary variable subscription source."""
    source = subscription_request.planetary_variable_source(
        var_type,
        var_id,
        geometry=geom_geojson,
        start_time=datetime(2021, 3, 1),
        end_time=datetime(2021, 3, 2),
    )

    if var_type:
        assert source["type"] == var_type
    else:
        assert "type" not in source
    params = source["parameters"]
    assert params["id"] == var_id
    assert params["geometry"] == geom_geojson
    assert params["start_time"].startswith("2021-03-01")


@respx.mock
@pytest.mark.parametrize(
    # Test all the combinations of the three options plus some with dupes.
    "publishing_stages",
    list(
        itertools.chain.from_iterable(
            itertools.combinations(["preview", "standard", "finalized"], i)
            for i in range(1, 4))) + [("preview", "preview"),
                                      ("preview", "finalized", "preview")])
def test_catalog_source_publishing_stages(publishing_stages,
                                          geom_geojson,
                                          mock_bundles):
    """Configure publishing stages for a catalog source."""
    source = subscription_request.catalog_source(
        item_types=["PSScene"],
        asset_types=["ortho_analytic_4b"],
        geometry=geom_geojson,
        start_time=datetime(2021, 3, 1),
        end_time=datetime(2023, 11, 1),
        rrule="FREQ=MONTHLY;BYMONTH=3,4,5,6,7,8,9,10",
        publishing_stages=publishing_stages,
    )

    assert source["parameters"]["publishing_stages"] == list(
        set(publishing_stages))


@respx.mock
def test_catalog_source_time_range_type_acquired(geom_geojson, mock_bundles):
    """Configure 'acquired' time range type for a catalog source."""
    source = subscription_request.catalog_source(
        item_types=["PSScene"],
        asset_types=["ortho_analytic_4b"],
        start_time=datetime(2021, 3, 1),
        time_range_type="acquired",
        geometry=geom_geojson,
    )

    assert source["parameters"]["time_range_type"] == "acquired"


def test_cloud_filter_tool_success():
    res = subscription_request.cloud_filter_tool(
        clear_percent=subscription_request.FilterValue(gte=90),
        cloud_percent=subscription_request.FilterValue(lte=10, gte=5))
    expected = {
        "type": "cloud_filter",
        "parameters": {
            "clear_percent": {
                "gte": 90
            },
            "cloud_percent": {
                "lte": 10, "gte": 5
            }
        }
    }

    assert res == expected



"""Planet Subscriptions API Python client."""

from typing import Any, Dict, Iterator, Optional, Sequence, Union

from typing_extensions import Literal

from planet.http import Session
from planet.clients import SubscriptionsClient


class SubscriptionsAPI:
    """Subscriptions API client

    Example:
        ```python
        >>> from planet import Planet
        >>>
        >>> pl = Planet()
        >>> pl.subscriptions.list_subscriptions()
        ```
    """

    _client: SubscriptionsClient

    def __init__(self,
                 session: Session,
                 base_url: Optional[str] = None) -> None:
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to production subscriptions
                API base url.
        """

        self._client = SubscriptionsClient(session, base_url)

    def list_subscriptions(self,
                           status: Optional[Sequence[str]] = None,
                           limit: int = 100,
                           created: Optional[str] = None,
                           end_time: Optional[str] = None,
                           hosting: Optional[bool] = None,
                           name__contains: Optional[str] = None,
                           name: Optional[str] = None,
                           source_type: Optional[str] = None,
                           start_time: Optional[str] = None,
                           sort_by: Optional[str] = None,
                           updated: Optional[str] = None,
                           page_size: int = 500) -> Iterator[dict]:
        """Iterate over list of account subscriptions with optional filtering.

        Note:
            The name of this method is based on the API's method name.
            This method provides iteration over subscriptions, it does
            not return a list.

        Args:
            created (str): filter by created time or interval.
            end_time (str): filter by end time or interval.
            hosting (bool): only return subscriptions that contain a
                hosting block (e.g. SentinelHub hosting).
            name__contains (str): only return subscriptions with a
                name that contains the given string.
            name (str): filter by name.
            source_type (str): filter by source type.
            start_time (str): filter by start time or interval.
            status (Set[str]): include subscriptions with a status in this set.
            sort_by (str): fields to sort subscriptions by. Multiple
                fields can be specified, separated by commas. The sort
                direction can be specified by appending ' ASC' or '
                DESC' to the field name. The default sort direction is
                ascending. When multiple fields are specified, the sort
                order is applied in the order the fields are listed.

                Supported fields: name, created, updated, start_time, end_time

                Examples:
                 * "name"
                 * "name DESC"
                 * "name,end_time DESC,start_time"
            updated (str): filter by updated time or interval.
            limit (int): limit the number of subscriptions in the
                results. When set to 0, no maximum is applied.
            page_size (int): number of subscriptions to return per page.
            TODO: user_id

        Datetime args (created, end_time, start_time, updated) can either be a
        date-time or an interval, open or closed. Date and time expressions adhere
        to RFC 3339. Open intervals are expressed using double-dots.

        Examples:
            * A date-time: "2018-02-12T23:20:50Z"
            * A closed interval: "2018-02-12T00:00:00Z/2018-03-18T12:31:12Z"
            * Open intervals: "2018-02-12T00:00:00Z/.." or "../2018-03-18T12:31:12Z"

        Yields:
            dict: a description of a subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """

        results = self._client.list_subscriptions(status,
                                                  limit,
                                                  created,
                                                  end_time,
                                                  hosting,
                                                  name__contains,
                                                  name,
                                                  source_type,
                                                  start_time,
                                                  sort_by,
                                                  updated,
                                                  page_size)

        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass

    def create_subscription(self, request: Dict) -> Dict:
        """Create a Subscription.

        Args:
            request (dict): description of a subscription.

        Returns:
            dict: description of created subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        return self._client._call_sync(
            self._client.create_subscription(request))

    def cancel_subscription(self, subscription_id: str) -> None:
        """Cancel a Subscription.

        Args:
            subscription_id (str): id of subscription to cancel.

        Returns:
            None

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        return self._client._call_sync(
            self._client.cancel_subscription(subscription_id))

    def update_subscription(self, subscription_id: str, request: dict) -> dict:
        """Update (edit) a Subscription via PUT.

        Args
            subscription_id (str): id of the subscription to update.
            request (dict): subscription content for update, full
                payload is required.

        Returns:
            dict: description of the updated subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        return self._client._call_sync(
            self._client.update_subscription(subscription_id, request))

    def patch_subscription(self, subscription_id: str,
                           request: Dict[str, Any]) -> Dict[str, Any]:
        """Update (edit) a Subscription via PATCH.

        Args
            subscription_id (str): id of the subscription to update.
            request (dict): subscription content for update, only
                attributes to update are required.

        Returns:
            dict: description of the updated subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        return self._client._call_sync(
            self._client.patch_subscription(subscription_id, request))

    def get_subscription(self, subscription_id: str) -> Dict[str, Any]:
        """Get a description of a Subscription.

        Args:
            subscription_id (str): id of a subscription.

        Returns:
            dict: description of the subscription.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        return self._client._call_sync(
            self._client.get_subscription(subscription_id))

    def get_results(
        self,
        subscription_id: str,
        status: Optional[Sequence[Literal["created",
                                          "queued",
                                          "processing",
                                          "failed",
                                          "success"]]] = None,
        limit: int = 100,
    ) -> Iterator[Union[Dict[str, Any], str]]:
        """Iterate over results of a Subscription.

        Notes:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            subscription_id (str): id of a subscription.
            status (Set[str]): pass result with status in this set,
                filter out results with status not in this set.
            limit (int): limit the number of subscriptions in the
                results. When set to 0, no maximum is applied.
            TODO: created, updated, completed, user_id

        Yields:
            dict: description of a subscription results.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        results = self._client.get_results(subscription_id, status, limit)

        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass

    def get_results_csv(
        self,
        subscription_id: str,
        status: Optional[Sequence[Literal["created",
                                          "queued",
                                          "processing",
                                          "failed",
                                          "success"]]] = None
    ) -> Iterator[str]:
        """Iterate over rows of results CSV for a Subscription.

        Parameters:
            subscription_id (str): id of a subscription.
            status (Set[str]): pass result with status in this set,
                filter out results with status not in this set.
            TODO: created, updated, completed, user_id

        Yields:
            str: a row from a CSV file.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        results = self._client.get_results_csv(subscription_id, status)
        # Note: retries are not implemented yet. This project has
        # retry logic for HTTP requests, but does not handle errors
        # during streaming. We may want to consider a retry decorator
        # for this entire method a la stamina:
        # https://github.com/hynek/stamina.
        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass

    def get_summary(self) -> Dict[str, Any]:
        """Summarize the status of all subscriptions via GET.

        Returns:
            dict: subscription totals by status.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        return self._client._call_sync(self._client.get_summary())

    def get_subscription_summary(self, subscription_id: str) -> Dict[str, Any]:
        """Summarize the status of results for a single subscription via GET.

        Args
            subscription_id (str): ID of the subscription to summarize.

        Returns:
            dict: result totals for the provided subscription by status.

        Raises:
            APIError: on an API server error.
            ClientError: on a client error.
        """
        return self._client._call_sync(
            self._client.get_subscription_summary(subscription_id))



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import respx
import pytest

from planet import exceptions, order_request, specs

LOGGER = logging.getLogger(__name__)

TEST_ID = 'doesntmatter'
TEST_PRODUCT_BUNDLE = 'analytic_sr'
TEST_FALLBACK_BUNDLE = 'analytic'
TEST_ITEM_TYPE = 'SkySatScene'
TEST_ARCHIVE_FILENAME = '{{name}}_b_{order_id}}.zip'


@respx.mock
def test_build_request(mock_bundles):
    product = {
        "item_ids": [TEST_ID],
        "item_type": TEST_ITEM_TYPE,
        "product_bundle": f'{TEST_PRODUCT_BUNDLE},{TEST_FALLBACK_BUNDLE}'
    }
    subscription_id = 5
    delivery = {
        'archive_type': 'zip',
        'single_archive': True,
        'archive_filename': TEST_ARCHIVE_FILENAME,
        'amazon_s3': {
            'aws_access_key_id': 'aws_access_key_id',
            'aws_secret_access_key': 'aws_secret_access_key',
            'bucket': 'bucket',
            'aws_region': 'aws_region'
        }
    }
    notifications = {
        'email': 'email', 'webhook': {
            'url': 'webhookurl', 'per_order': True
        }
    }
    order_type = 'partial'
    tool = {'bandmath': 'jsonstring'}
    stac_json = {'stac': {}}

    request = order_request.build_request('test_name', [product],
                                          subscription_id=subscription_id,
                                          delivery=delivery,
                                          notifications=notifications,
                                          order_type=order_type,
                                          tools=[tool],
                                          stac=stac_json)
    expected = {
        'name': 'test_name',
        'products': [product],
        'subscription_id': subscription_id,
        'delivery': delivery,
        'notifications': notifications,
        'order_type': order_type,
        'tools': [tool],
        'metadata': stac_json
    }
    assert request == expected

    order_type = 'notsupported'
    with pytest.raises(specs.SpecificationException):
        _ = order_request.build_request('test_name', [product],
                                        subscription_id=subscription_id,
                                        delivery=delivery,
                                        notifications=notifications,
                                        order_type=order_type,
                                        tools=[tool],
                                        stac=stac_json)


@respx.mock
def test_product(mock_bundles):
    product_config = order_request.product(
        [TEST_ID],
        TEST_PRODUCT_BUNDLE,
        TEST_ITEM_TYPE,
        fallback_bundle=TEST_FALLBACK_BUNDLE)

    expected = {
        "item_ids": [TEST_ID],
        "item_type": TEST_ITEM_TYPE,
        "product_bundle": f'{TEST_PRODUCT_BUNDLE},{TEST_FALLBACK_BUNDLE}'
    }
    assert product_config == expected

    with pytest.raises(specs.SpecificationException):
        _ = order_request.product([TEST_ID],
                                  'notsupported',
                                  TEST_ITEM_TYPE,
                                  fallback_bundle=TEST_FALLBACK_BUNDLE)

    with pytest.raises(specs.SpecificationException):
        _ = order_request.product([TEST_ID],
                                  TEST_PRODUCT_BUNDLE,
                                  'notsupported',
                                  fallback_bundle=TEST_FALLBACK_BUNDLE)

    with pytest.raises(specs.SpecificationException):
        _ = order_request.product([TEST_ID],
                                  TEST_PRODUCT_BUNDLE,
                                  TEST_ITEM_TYPE,
                                  fallback_bundle='notsupported')


def test_notifications():
    notifications_config = order_request.notifications(
        email=True, webhook_url='webhookurl', webhook_per_order=True)
    expected = {
        'email': True, 'webhook': {
            'url': 'webhookurl', 'per_order': True
        }
    }
    assert notifications_config == expected

    empty_notifications_config = order_request.notifications()
    empty_expected = {}
    assert empty_notifications_config == empty_expected


def test_delivery():
    as3_config = {
        'amazon_s3': {
            'aws_access_key_id': 'aws_access_key_id',
            'aws_secret_access_key': 'aws_secret_access_key',
            'bucket': 'bucket',
            'aws_region': 'aws_region'
        }
    }
    delivery_config = order_request.delivery('zip',
                                             True,
                                             TEST_ARCHIVE_FILENAME,
                                             cloud_config=as3_config)

    expected = {
        'archive_type': 'zip',
        'single_archive': True,
        'archive_filename': TEST_ARCHIVE_FILENAME,
        'amazon_s3': {
            'aws_access_key_id': 'aws_access_key_id',
            'aws_secret_access_key': 'aws_secret_access_key',
            'bucket': 'bucket',
            'aws_region': 'aws_region'
        }
    }
    assert delivery_config == expected


def test_delivery_missing_archive_details():
    as3_config = {
        'amazon_s3': {
            'aws_access_key_id': 'aws_access_key_id',
            'aws_secret_access_key': 'aws_secret_access_key',
            'bucket': 'bucket',
            'aws_region': 'aws_region'
        }
    }
    delivery_config = order_request.delivery(archive_type='zip',
                                             cloud_config=as3_config)

    expected = {
        'archive_type': 'zip',
        'archive_filename': "{{name}}_{{order_id}}.zip",
        'single_archive': False,
        'amazon_s3': {
            'aws_access_key_id': 'aws_access_key_id',
            'aws_secret_access_key': 'aws_secret_access_key',
            'bucket': 'bucket',
            'aws_region': 'aws_region'
        }
    }
    assert delivery_config == expected


def test_amazon_s3():
    as3_config = order_request.amazon_s3('aws_access_key_id',
                                         'aws_secret_access_key',
                                         'bucket',
                                         'aws_region')
    expected = {
        'amazon_s3': {
            'aws_access_key_id': 'aws_access_key_id',
            'aws_secret_access_key': 'aws_secret_access_key',
            'bucket': 'bucket',
            'aws_region': 'aws_region'
        }
    }
    assert as3_config == expected


def test_azure_blob_storage():
    abs_config = order_request.azure_blob_storage('account',
                                                  'container',
                                                  'sas_token')
    expected = {
        'azure_blob_storage': {
            'account': 'account',
            'container': 'container',
            'sas_token': 'sas_token',
        }
    }
    assert abs_config == expected


def test_google_cloud_storage():
    gcs_config = order_request.google_cloud_storage('bucket', 'credentials')

    expected = {
        'google_cloud_storage': {
            'bucket': 'bucket',
            'credentials': 'credentials',
        }
    }
    assert gcs_config == expected


def test_google_earth_engine():
    gee_config = order_request.google_earth_engine('project', 'collection')
    expected = {
        'google_earth_engine': {
            'project': 'project',
            'collection': 'collection',
        }
    }

    assert gee_config == expected


def test_oracle_cloud_storage():
    ocs_config = order_request.oracle_cloud_storage('ocs_access_key_id',
                                                    'ocs_secret_access_key',
                                                    'bucket',
                                                    'region',
                                                    'namespace')

    expected = {
        'oracle_cloud_storage': {
            'customer_access_key_id': 'ocs_access_key_id',
            'customer_secret_key': 'ocs_secret_access_key',
            'bucket': 'bucket',
            'region': 'region',
            'namespace': 'namespace'
        }
    }
    assert ocs_config == expected


def test__tool():
    test_tool = order_request._tool('bandmath', 'jsonstring')
    assert test_tool == {'bandmath': 'jsonstring'}

    with pytest.raises(specs.SpecificationException):
        _ = order_request._tool('notsupported', 'jsonstring')


def test_clip_tool_polygon(geom_geojson):
    ct = order_request.clip_tool(geom_geojson)
    expected = {'clip': {'aoi': geom_geojson}}
    assert ct == expected


def test_clip_tool_multipolygon(multipolygon_geom_geojson):
    ct = order_request.clip_tool(multipolygon_geom_geojson)
    expected = {'clip': {'aoi': multipolygon_geom_geojson}}
    assert ct == expected


def test_clip_tool_invalid(point_geom_geojson):
    """Confirm an exception is raised if an invalid geometry type is supplied.
    """
    with pytest.raises(exceptions.ClientError):
        order_request.clip_tool(point_geom_geojson)


def test_composite_tool_no_group_by():
    composite_tool = order_request.composite_tool()
    expected = {'composite': {}}
    assert composite_tool == expected


def test_composite_tool_group_by_strip_id():
    composite_tool = order_request.composite_tool(group_by='strip_id')
    expected = {'composite': {'group_by': 'strip_id'}}
    assert composite_tool == expected


def test_composite_tool_group_by_order():
    composite_tool = order_request.composite_tool(group_by='order')
    expected = {'composite': {'group_by': 'order'}}
    assert composite_tool == expected


def test_composite_tool_group_by_invalid():
    with pytest.raises(exceptions.ClientError):
        order_request.composite_tool(group_by='invalid')


def test_reproject_tool():
    rt = order_request.reproject_tool(projection='proj', resolution=5)
    expected = {'reproject': {'projection': 'proj', 'resolution': 5}}
    assert rt == expected


def test_tile_tool():
    tt = order_request.tile_tool(30, pixel_size=3)
    expected = {'tile': {'tile_size': 30, 'pixel_size': 3}}
    assert tt == expected


def test_toar_tool():
    tt = order_request.toar_tool(scale_factor=5)
    expected = {'toar': {'scale_factor': 5}}
    assert tt == expected

    tt_empty = order_request.toar_tool()
    expected_empty = {'toar': {}}
    assert tt_empty == expected_empty


def test_harmonization_tool_success():
    ht = order_request.harmonize_tool("Sentinel-2")
    expected = {'harmonize': {'target_sensor': "Sentinel-2"}}
    assert ht == expected


def test_harmonization_tool_invalid_target_sensor():
    with pytest.raises(exceptions.ClientError):
        order_request.harmonize_tool('invalid')


def test_band_math_tool_success():
    res = order_request.band_math_tool(b1='b1', b2='arctan(b1)')

    expected = {
        "bandmath": {
            "b1": "b1",
            "b2": "arctan(b1)",
            "pixel_type": specs.BAND_MATH_PIXEL_TYPE_DEFAULT
        }
    }
    assert res == expected


def test_band_math_tool_invalid_pixel_type():
    with pytest.raises(exceptions.ClientError):
        order_request.band_math_tool(b1='b1',
                                     b2='arctan(b1)',
                                     pixel_type="invalid")


def test_no_archive_items_without_type():
    """Without an archive type no filename or single option are passed."""
    delivery_config = order_request.delivery(
        None, True, TEST_ARCHIVE_FILENAME, cloud_config={"bogus_storage": {}})

    assert "bogus_storage" in delivery_config
    assert "archive_type" not in delivery_config
    assert "archive_filename" not in delivery_config
    assert "single_archive" not in delivery_config


def test_sentinel_hub():
    sh_config = order_request.sentinel_hub()
    expected = {'sentinel_hub': {}}
    assert sh_config == expected


def test_sentinel_hub_collection_id():
    sh_config = order_request.sentinel_hub("1234")
    expected = {'sentinel_hub': {'collection_id': "1234"}}
    assert sh_config == expected


def test_sentinel_hub_create_configuration():
    sh_config = order_request.sentinel_hub(create_configuration=True)
    expected = {'sentinel_hub': {'create_configuration': True}}
    assert sh_config == expected


def test_sentinel_hub_collection_configuration():
    # Note, this behavior will be rejected by the API, but it is valid in building a request
    sh_config = order_request.sentinel_hub(collection_id="1234",
                                           create_configuration=True)
    expected = {
        'sentinel_hub': {
            'collection_id': '1234', 'create_configuration': True
        }
    }
    assert sh_config == expected


@pytest.mark.parametrize(
    "bundle, fallback_bundle",
    [
        ('analytic_8b_udm2', 'analytic_udm2'),
        ('analytic_8b_udm2', 'analytic_udm2,analytic_3b_udm2'),
        ('analytic_8b_udm2', ['analytic_udm2', 'analytic_3b_udm2']),
    ])
def test_fallback_bundle_success(bundle, fallback_bundle):
    product = order_request.product(["20250130_035211_69_2516"],
                                    bundle,
                                    "PSScene",
                                    fallback_bundle=fallback_bundle)

    # Ensure fallback_bundle is a string for the expected output
    fallback_bundle = fallback_bundle if isinstance(
        fallback_bundle, str) else ','.join(fallback_bundle)

    expected = {
        "item_ids": ["20250130_035211_69_2516"],
        "item_type": "PSScene",
        "product_bundle": f"{bundle},{fallback_bundle}"
    }
    assert product == expected


@pytest.mark.parametrize(
    "bundle, fallback_bundle",
    [
        ('analytic_8b_udm2', ''),
        ('analytic_8b_udm2', 'analytic_udm2;analytic_3b_udm2'),
        ('analytic_8b_udm2', 'analytic_udm2,analytic_3b_udm2,fake_bundle'),
        ('analytic_8b_udm2',
         ['analytic_udm2', 'analytic_3b_udm2', 'fake_bundle']),
    ])
def test_fallback_bundle_invalid(bundle, fallback_bundle):
    with pytest.raises(specs.SpecificationException):
        _ = order_request.product(["20250130_035211_69_2516"],
                                  bundle,
                                  "PSScene",
                                  fallback_bundle=fallback_bundle)



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import base64
from http import HTTPStatus
import json

import httpx
import respx

import pytest

# from planet.auth import _SecretFile
from planet import auth
from planet.cli import session
from planet.exceptions import AuthException

TEST_URL = 'mock://mock.com'


# skip the global mock of _SecretFile.read
# for this module
@pytest.fixture(autouse=True, scope='module')
def test_secretfile_read():
    return


@pytest.fixture()
def test_valid_secretfile(tmp_path, monkeypatch):
    # write our own secret file
    secret_path = f'{tmp_path}/secret.test'
    monkeypatch.setattr(auth, 'SECRET_FILE_PATH', secret_path)
    with open(secret_path, 'w') as fp:
        json.dump({'key': 'clisessiontest'}, fp)


@respx.mock
@pytest.mark.anyio
async def test_CliSession_headers(test_valid_secretfile):
    async with session.CliSession() as sess:
        route = respx.get(TEST_URL)
        route.return_value = httpx.Response(HTTPStatus.OK)

        await sess.request(method='GET', url=TEST_URL)

        # the proper headers are included and they have the expected values
        received_request = route.calls.last.request
        assert received_request.headers['x-planet-app'] == 'python-cli'
        assert 'planet-client-python/' in received_request.headers[
            'user-agent']


@respx.mock
@pytest.mark.anyio
async def test_CliSession_auth_valid(test_valid_secretfile):
    async with session.CliSession() as sess:
        route = respx.get(TEST_URL)
        route.return_value = httpx.Response(HTTPStatus.OK)

        await sess.request(method='GET', url=TEST_URL)

        # the proper headers are included and they have the expected values
        received_request = route.calls.last.request
        credentials = received_request.headers['authorization'].strip(
            'Authorization: Basic ')
        assert base64.b64decode(credentials) == b'clisessiontest:'


@respx.mock
@pytest.mark.anyio
async def test_CliSession_auth_invalid(tmp_path, monkeypatch):
    # write invalid secret file
    secret_path = f'{tmp_path}/secret.test'
    monkeypatch.setattr(auth, 'SECRET_FILE_PATH', secret_path)
    with open(secret_path, 'w') as fp:
        json.dump({'invalidkey': 'clisessiontest'}, fp)

    with pytest.raises(AuthException):
        session.CliSession()


@respx.mock
@pytest.mark.anyio
async def test_CliSession_auth_nofile(tmp_path, monkeypatch):
    # point to non-existant file
    secret_path = f'{tmp_path}/doesnotexist.test'
    monkeypatch.setattr(auth, 'SECRET_FILE_PATH', secret_path)

    with pytest.raises(AuthException):
        session.CliSession()



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Auth API CLI"""
import logging
import os

import click

import planet
from planet.constants import ENV_API_KEY
from .cmds import translate_exceptions

LOGGER = logging.getLogger(__name__)


@click.group()  # type: ignore
@click.pass_context
@click.option('-u',
              '--base-url',
              default=None,
              help='Assign custom base Auth API URL.')
def auth(ctx, base_url):
    """Commands for working with Planet authentication"""
    ctx.obj['BASE_URL'] = base_url


@auth.command()  # type: ignore
@click.pass_context
@translate_exceptions
@click.option(
    '--email',
    default=None,
    prompt=True,
    help=('The email address associated with your Planet credentials.'))
@click.password_option('--password',
                       confirmation_prompt=False,
                       help=('Account password. Will not be saved.'))
def init(ctx, email, password):
    """Obtain and store authentication information"""
    base_url = ctx.obj['BASE_URL']
    plauth = planet.Auth.from_login(email, password, base_url=base_url)
    plauth.store()
    click.echo('Initialized')
    if os.getenv(ENV_API_KEY):
        click.echo(f'Warning - Environment variable {ENV_API_KEY} already '
                   'exists. To update, with the new value, use the following:')
        click.echo(f'export {ENV_API_KEY}=$(planet auth value)')


@auth.command()  # type: ignore
@translate_exceptions
def value():
    """Print the stored authentication information"""
    click.echo(planet.Auth.from_file().value)


@auth.command()  # type: ignore
@translate_exceptions
@click.argument('key')
def store(key):
    """Store authentication information"""
    plauth = planet.Auth.from_key(key)
    if click.confirm('This overrides the stored value. Continue?'):
        plauth.store()
        click.echo('Updated')
        if os.getenv(ENV_API_KEY):
            click.echo(f'Warning - Environment variable {ENV_API_KEY} already '
                       'exists. To update, with the new value, use the '
                       'following:')
            click.echo(f'export {ENV_API_KEY}=$(planet auth value)')



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for interacting with the data api"""
import asyncio
import hashlib
import logging
from pathlib import Path
import time
from typing import Any, AsyncIterator, Awaitable, Callable, Dict, List, Optional, TypeVar, Union
import uuid

from ..data_filter import empty_filter
from .. import exceptions
from ..constants import PLANET_BASE_URL
from ..http import Session
from ..models import GeojsonLike, Paged, StreamingBody
from ..specs import validate_data_item_type
from ..geojson import as_geom_or_ref

BASE_URL = f'{PLANET_BASE_URL}/data/v1/'
SEARCHES_PATH = '/searches'
STATS_PATH = '/stats'

# TODO: get these values from the spec directly gh-619
# NOTE: these values are mached against a lower-case value so must also be
# lower-case
LIST_SORT_ORDER = ('created desc', 'created asc')
LIST_SEARCH_TYPE = ('any', 'saved', 'quick')
LIST_SORT_DEFAULT = 'created desc'
LIST_SEARCH_TYPE_DEFAULT = 'any'

SEARCH_SORT = ('published desc',
               'published asc',
               'acquired desc',
               'acquired asc')
SEARCH_SORT_DEFAULT = 'published desc'
STATS_INTERVAL = ('hour', 'day', 'week', 'month', 'year')

WAIT_DELAY = 5
WAIT_MAX_ATTEMPTS = 200

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


class Items(Paged):
    """Asynchronous iterator over items from a paged response."""
    NEXT_KEY = '_next'
    ITEMS_KEY = 'features'


class Searches(Paged):
    """Asynchronous iterator over searches from a paged response."""
    NEXT_KEY = '_next'
    ITEMS_KEY = 'searches'


class DataClient:
    """High-level asynchronous access to Planet's data API.

    Example:
        ```python
        >>> import asyncio
        >>> from planet import Session
        >>>
        >>> async def main():
        ...     async with Session() as sess:
        ...         cl = sess.client('data')
        ...         # use client here
        ...
        >>> asyncio.run(main())

        ```
    """

    def __init__(self, session: Session, base_url: Optional[str] = None):
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to production data API
                base url.
        """
        self._session = session

        self._base_url = base_url or BASE_URL
        if self._base_url.endswith('/'):
            self._base_url = self._base_url[:-1]

    def _call_sync(self, f: Awaitable[T]) -> T:
        """block on an async function call, using the call_sync method of the session"""
        return self._session._call_sync(f)

    @staticmethod
    def _check_search_id(sid):
        """Raises planet.exceptions.ClientError if sid is not a valid UUID"""
        try:
            uuid.UUID(hex=sid)
        except (ValueError, AttributeError):
            msg = f'Search id ({sid}) is not a valid UUID hexadecimal string.'
            raise exceptions.ClientError(msg)

    def _searches_url(self):
        return f'{self._base_url}{SEARCHES_PATH}'

    def _item_url(self, item_type, item_id):
        return f'{self._base_url}/item-types/{item_type}/items/{item_id}'

    async def search(
            self,
            item_types: List[str],
            search_filter: Optional[dict] = None,
            name: Optional[str] = None,
            sort: Optional[str] = None,
            limit: int = 100,
            geometry: Optional[GeojsonLike] = None) -> AsyncIterator[dict]:
        """Iterate over results from a quick search.

        Quick searches are saved for a short period of time (~month). The
        `name` parameter of the search defaults to the id of the generated
        search id if `name` is not specified.

        Note:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            item_types: The item types to include in the search.
            search_filter: Structured search criteria to apply. If None,
                no search criteria is applied.
            sort: Field and direction to order results by. Valid options are
            given in SEARCH_SORT.
            name: The name of the saved search.
            limit: Maximum number of results to return. When set to 0, no
                maximum is applied.
            geometry: GeoJSON, a feature reference or a list of feature
                references

        Yields:
            Description of an item.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        url = f'{self._base_url}/quick-search'

        search_filter = search_filter or empty_filter()

        item_types = [validate_data_item_type(item) for item in item_types]
        request_json = {'filter': search_filter, 'item_types': item_types}

        if geometry:
            request_json['geometry'] = as_geom_or_ref(geometry)
        if name:
            request_json['name'] = name

        params = {}
        if sort and sort != SEARCH_SORT_DEFAULT:
            sort = sort.lower()
            if sort not in SEARCH_SORT:
                raise exceptions.ClientError(
                    f'{sort} must be one of {SEARCH_SORT}')
            params['_sort'] = sort
        response = await self._session.request(method='POST',
                                               url=url,
                                               json=request_json,
                                               params=params)
        async for i in Items(response, self._session.request, limit=limit):
            yield i

    async def create_search(
        self,
        item_types: List[str],
        search_filter: dict,
        name: str,
        enable_email: bool = False,
        geometry: Optional[Union[dict, str]] = None,
    ) -> dict:
        """Create a new saved structured item search.

        To filter to items you have access to download which are of standard
        (aka not test) quality, use the following:

        ```python
        >>> from planet import data_filter
        >>> data_filter.and_filter([
        ...     data_filter.permission_filter(),
        ...     data_filter.std_quality_filter()
        >>> ])

        ```

        To avoid filtering out any imagery, supply a blank AndFilter, which can
        be created with `data_filter.and_filter([])`.


        Parameters:
            item_types: The item types to include in the search.
            search_filter: Structured search criteria.
            name: The name of the saved search.
            enable_email: Send a daily email when new results are added.
            geometry: A feature reference or a GeoJSON

        Returns:
            Description of the saved search.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        url = self._searches_url()

        item_types = [validate_data_item_type(item) for item in item_types]

        request = {
            'name': name,
            'filter': search_filter,
            'item_types': item_types,
            '__daily_email_enabled': enable_email
        }
        if geometry:
            request['geometry'] = as_geom_or_ref(geometry)

        response = await self._session.request(method='POST',
                                               url=url,
                                               json=request)
        return response.json()

    async def update_search(
            self,
            search_id: str,
            item_types: List[str],
            search_filter: dict,
            name: str,
            enable_email: bool = False,
            geometry: Optional[Union[dict, str]] = None) -> dict:
        """Update an existing saved search.

        Parameters:
            search_id: Saved search identifier.
            item_types: The item types to include in the search.
            search_filter: Structured search criteria.
            name: The name of the saved search.
            enable_email: Send a daily email when new results are added.
            geometry: A feature reference or a GeoJSON

        Returns:
            Description of the saved search.
        """
        url = f'{self._searches_url()}/{search_id}'

        item_types = [validate_data_item_type(item) for item in item_types]

        request = {
            'name': name,
            'filter': search_filter,
            'item_types': item_types,
            '__daily_email_enabled': enable_email
        }
        if geometry:
            request['geometry'] = geometry

        response = await self._session.request(method='PUT',
                                               url=url,
                                               json=request)
        return response.json()

    async def list_searches(self,
                            sort: str = LIST_SORT_DEFAULT,
                            search_type: str = LIST_SEARCH_TYPE_DEFAULT,
                            limit: int = 100) -> AsyncIterator[dict]:
        """Iterate through list of searches available to the user.

        Note:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            sort: Field and direction to order results by.
            search_type: Filter to specified search type.
            limit: Maximum number of results to return. When set to 0, no
                maximum is applied.

        Yields:
            Description of a search.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If sort or search_type are not
                valid.
        """
        params = {}

        sort = sort.lower()
        if sort not in LIST_SORT_ORDER:
            raise exceptions.ClientError(
                f'{sort} must be one of {LIST_SORT_ORDER}')
        elif sort != LIST_SORT_DEFAULT:
            params['_sort'] = sort

        search_type = search_type.lower()
        if search_type not in LIST_SEARCH_TYPE:
            raise exceptions.ClientError(
                f'{search_type} must be one of {LIST_SEARCH_TYPE}')
        elif search_type != LIST_SEARCH_TYPE_DEFAULT:
            params['search_type'] = search_type

        url = self._searches_url()
        response = await self._session.request(method='GET',
                                               url=url,
                                               params=params)
        async for s in Searches(response, self._session.request, limit=limit):
            yield s

    async def delete_search(self, search_id: str):
        """Delete an existing saved search.

        Parameters:
            search_id: Saved search identifier.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        url = f'{self._searches_url()}/{search_id}'

        await self._session.request(method='DELETE', url=url)

    async def get_search(self, search_id: str) -> dict:
        """Get a saved search by id.

        Parameters:
            search_id: Stored search identifier.

        Returns:
            Saved search details.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        url = f'{self._searches_url()}/{search_id}'

        response = await self._session.request(method='GET', url=url)
        return response.json()

    async def run_search(self,
                         search_id: str,
                         sort: Optional[str] = None,
                         limit: int = 100) -> AsyncIterator[dict]:
        """Iterate over results from a saved search.

        Note:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            search_id: Stored search identifier.
            sort: Field and direction to order results by. Valid options are
            given in SEARCH_SORT.
            limit: Maximum number of results to return. When set to 0, no
                maximum is applied.

        Yields:
            Description of an item.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If search_id or sort is not valid.
        """
        self._check_search_id(search_id)
        url = f'{self._searches_url()}/{search_id}/results'

        params = {}
        if sort and sort != SEARCH_SORT_DEFAULT:
            sort = sort.lower()
            if sort not in SEARCH_SORT:
                raise exceptions.ClientError(
                    f'{sort} must be one of {SEARCH_SORT}')
            params['_sort'] = sort

        response = await self._session.request(method='GET',
                                               url=url,
                                               params=params)
        async for i in Items(response, self._session.request, limit=limit):
            yield i

    async def get_stats(self,
                        item_types: List[str],
                        search_filter: dict,
                        interval: str) -> dict:
        """Get item search statistics.

        Parameters:
            item_types: The item types to include in the search.
            search_filter: Structured search criteria.
            interval: The size of the histogram date buckets.

        Returns:
            A full JSON description of the returned statistics result
            histogram.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If interval is not valid.
        """
        interval = interval.lower()
        if interval not in STATS_INTERVAL:
            raise exceptions.ClientError(
                f'{interval} must be one of {STATS_INTERVAL}')

        url = f'{self._base_url}{STATS_PATH}'

        item_types = [validate_data_item_type(item) for item in item_types]
        request = {
            'interval': interval,
            'filter': search_filter,
            'item_types': item_types
        }

        response = await self._session.request(method='POST',
                                               url=url,
                                               json=request)
        return response.json()

    async def list_item_assets(self, item_type_id: str, item_id: str) -> dict:
        """List all assets available for an item.

        An asset describes a product that can be derived from an item's source
        data, and can be used for various analytic, visual or other purposes.
        These are referred to as asset_types.

        Parameters:
            item_type_id: Item type identifier.
            item_id: Item identifier.

        Returns:
            Descriptions of available assets as a dictionary with asset_type_id
            as keys and asset description as value.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        url = f'{self._item_url(item_type_id, item_id)}/assets'

        response = await self._session.request(method='GET', url=url)
        return response.json()

    async def get_asset(self,
                        item_type_id: str,
                        item_id: str,
                        asset_type_id: str) -> dict:
        """Get an item asset description.

        Parameters:
            item_type_id: Item type identifier.
            item_id: Item identifier.
            asset_type_id: Asset type identifier.

        Returns:
            Description of the asset.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset type identifier is not
            valid.
        """
        item_type_id = validate_data_item_type(item_type_id)
        assets = await self.list_item_assets(item_type_id, item_id)

        try:
            asset = assets[asset_type_id]
        except KeyError:
            valid = list(assets.keys())
            raise exceptions.ClientError(
                f'asset_type_id ({asset_type_id}) must be one of {valid}')

        return asset

    async def activate_asset(self, asset: dict):
        """Activate an item asset.

        Parameters:
            asset: Description of the asset. Obtained from get_asset().

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset description is not
            valid.
        """
        try:
            status = asset['status']
        except KeyError:
            raise exceptions.ClientError('asset missing ["status"] entry.')

        try:
            url = asset['_links']['activate']
        except KeyError:
            raise exceptions.ClientError(
                'asset missing ["_links"]["activate"] entry')

        # lets not try to activate an asset already activating or active
        if status == 'inactive':
            # no response is returned
            await self._session.request(method='GET', url=url)

        return

    async def wait_asset(
            self,
            asset: dict,
            delay: int = WAIT_DELAY,
            max_attempts: int = WAIT_MAX_ATTEMPTS,
            callback: Optional[Callable[[str],
                                        None]] = None) -> Dict[Any, Any]:
        """Wait for an item asset to be active.

        Prior to waiting for the asset to be active, be sure to activate the
        asset with activate_asset().

        Parameters:
            asset: Description of the asset. Obtained from get_asset().
            delay: Time (in seconds) between polls.
            max_attempts: Maximum number of polls. When set to 0, no limit
                is applied.
            callback: Function that handles state progress updates.

        Returns:
            Last received description of the asset.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset_type_id is not valid or is
                not available or if the maximum number of attempts is reached
                before the asset is active.
        """
        # loop without end if max_attempts is zero
        # otherwise, loop until num_attempts reaches max_attempts
        num_attempts = 0
        while not max_attempts or num_attempts < max_attempts:
            t = time.time()

            try:
                current_status = asset['status']
            except KeyError:
                raise exceptions.ClientError('asset missing ["status"] entry.')

            LOGGER.debug(current_status)

            if callback:
                callback(current_status)

            if current_status == 'active':
                break

            sleep_time = max(delay - (time.time() - t), 0)
            LOGGER.debug(f'sleeping {sleep_time}s')
            await asyncio.sleep(sleep_time)

            num_attempts += 1

            try:
                asset_url = asset['_links']['_self']
            except KeyError:
                raise exceptions.ClientError(
                    'asset missing ["_links"]["_self"] entry.')

            response = await self._session.request(method='GET', url=asset_url)
            asset = response.json()

        if max_attempts and num_attempts >= max_attempts:
            raise exceptions.ClientError(
                f'Maximum number of attempts ({max_attempts}) reached.')

        return asset

    async def download_asset(self,
                             asset: dict,
                             filename: Optional[str] = None,
                             directory: Path = Path('.'),
                             overwrite: bool = False,
                             progress_bar: bool = True) -> Path:
        """Download an asset.

        The asset must be active before it can be downloaded. This can be
        achieved with activate_asset() followed by wait_asset().

        If overwrite is False and the file already exists, download will be
        skipped and the file path will be returned as usual.

        Parameters:
            asset: Description of the asset. Obtained from get_asset() or
                wait_asset().
            filename: Custom name to assign to downloaded file.
            directory: Base directory for file download.
            overwrite: Overwrite any existing files.
            progress_bar: Show progress bar during download.

        Returns:
            Path to downloaded file.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset is not active or asset
            description is not valid.
        """
        try:
            location = asset['location']
        except KeyError:
            raise exceptions.ClientError(
                'asset missing ["location"] entry. Is asset active?')

        async with self._session.stream(method='GET', url=location) as resp:
            body = StreamingBody(resp)
            dl_path = Path(directory, filename or body.name)
            dl_path.parent.mkdir(exist_ok=True, parents=True)
            await body.write(dl_path,
                             overwrite=overwrite,
                             progress_bar=progress_bar)
        return dl_path

    @staticmethod
    def validate_checksum(asset: dict, filename: Path):
        """Validate checksum of downloaded file

        Compares checksum calculated from the file against the value provided
        in the asset.

        Parameters:
            asset: Description of the asset. Obtained from get_asset() or
                wait_asset().
            filename: Full path to downloaded file.

        Raises:
            planet.exceptions.ClientError: If the file does not exist or if
                checksums do not match.
        """
        try:
            file_hash = hashlib.md5(filename.read_bytes()).hexdigest()
        except FileNotFoundError:
            raise exceptions.ClientError(f'File ({filename}) does not exist.')

        try:
            origin_hash = asset['md5_digest']
        except KeyError:
            raise exceptions.ClientError(
                'asset missing ["md5_digest"] entry. Is asset active?')

        if origin_hash != file_hash:
            raise exceptions.ClientError(
                f'File ({filename}) checksums do not match.')



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .http import Session
from . import data_filter, order_request, reporting, subscription_request
from .__version__ import __version__  # NOQA
from .auth import Auth
from .clients import DataClient, FeaturesClient, OrdersClient, SubscriptionsClient  # NOQA
from .io import collect
from .sync import Planet

__all__ = [
    'Auth',
    'collect',
    'DataClient',
    'data_filter',
    'FeaturesClient',
    'OrdersClient',
    'order_request',
    'Planet',
    'reporting',
    'Session',
    'SubscriptionsClient',
    'subscription_request'
]



# Copyright 2022 Planet Labs, PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for collecting a sequence into JSON."""
import json
import logging

import click

import planet
from .cmds import coro, translate_exceptions
from .io import echo_json
from .options import pretty

LOGGER = logging.getLogger(__name__)


@click.command()
@click.pass_context
@translate_exceptions
@coro
@click.argument('input', type=click.File('r'))
@pretty
async def collect(ctx, input, pretty):
    """Collect a sequence of JSON descriptions into a single JSON blob.

    If the descriptions represent GeoJSON features, a GeoJSON FeatureCollection
    is returned.

    Output can be pretty-printed with --pretty option.
    """

    # make an AsyncGenerator from the input lines
    async def _entries_aiter():
        for line in input:
            yield json.loads(line)

    entries = _entries_aiter()
    collected = await planet.collect(entries)
    echo_json(collected, pretty)



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import respx
import pytest

from planet import specs

LOGGER = logging.getLogger(__name__)

TEST_PRODUCT_BUNDLE = 'visual'
TEST_ITEM_TYPE = 'PSScene'
TEST_ASSET_TYPE = "basic_udm2"


def test_get_type_match():
    spec_list = ['Locket', 'drop', 'DEER']

    test_entry = 'locket'
    field_name = 'field_name'
    assert 'Locket' == specs.get_match(test_entry, spec_list, field_name)

    with pytest.raises(specs.SpecificationException):
        specs.get_match('a', ['b'], field_name)


@respx.mock
def test_validate_bundle_supported(mock_bundles):
    assert 'visual' == specs.validate_bundle(TEST_ITEM_TYPE, 'VISUAL')


@respx.mock
def test_validate_bundle_notsupported(mock_bundles):
    with pytest.raises(specs.SpecificationException):
        specs.validate_bundle(TEST_ITEM_TYPE, 'notsupported')


@respx.mock
def test_validate_bundle_notsupported_item_type(mock_bundles):
    with pytest.raises(specs.SpecificationException):
        specs.validate_item_type('wha')


@respx.mock
def test_validate_item_type_supported(mock_bundles):
    assert 'PSScene' == specs.validate_item_type('PSScene')


@respx.mock
def test_validate_item_type_notsupported_itemtype(mock_bundles):
    with pytest.raises(specs.SpecificationException):
        specs.validate_item_type('notsupported')


@respx.mock
def test_validate_data_item_type(mock_bundles):
    """ensure skysatvideo is included"""
    specs.validate_data_item_type('skysatvideo')


def test_validate_order_type_supported():
    assert 'full' == specs.validate_order_type('FULL')


def test_validate_order_type_notsupported():
    with pytest.raises(specs.SpecificationException):
        specs.validate_order_type('notsupported')


def test_validate_archive_type_supported():
    assert 'zip' == specs.validate_archive_type('ZIP')


def test_validate_archive_type_notsupported():
    with pytest.raises(specs.SpecificationException):
        specs.validate_archive_type('notsupported')


def test_validate_file_format_supported():
    assert 'COG' == specs.validate_file_format('cog')


def test_validate_file_format_notsupported():
    with pytest.raises(specs.SpecificationException):
        specs.validate_archive_type('notsupported')


@respx.mock
def test_get_product_bundles_with_item_type(mock_bundles):
    bundles = specs.get_product_bundles(item_type=TEST_ITEM_TYPE)
    assert TEST_PRODUCT_BUNDLE in bundles


@respx.mock
def test_get_product_bundles_without_item_type(mock_bundles):
    """assert an expected product bundle is in the list of all product bundles"""
    bundles = specs.get_product_bundles()
    assert TEST_PRODUCT_BUNDLE in bundles


@respx.mock
def test_get_item_types_with_bundle(mock_bundles):
    item_types = specs.get_item_types(product_bundle=TEST_PRODUCT_BUNDLE)
    assert TEST_ITEM_TYPE in item_types


@respx.mock
def test_get_item_types_without_bundle(mock_bundles):
    item_types = specs.get_item_types()
    assert TEST_ITEM_TYPE in item_types


@respx.mock
def test_validate_supported_bundles_success(mock_bundles):
    validated_bundle = specs.validate_supported_bundles(
        TEST_ITEM_TYPE,
        TEST_PRODUCT_BUNDLE,
    )
    assert validated_bundle == TEST_PRODUCT_BUNDLE


@respx.mock
def test_validate_supported_bundles_fail(mock_bundles):
    with pytest.raises(specs.SpecificationException):
        specs.validate_supported_bundles(
            TEST_ITEM_TYPE,
            'analytic',
        )


@respx.mock
def test_get_supported_assets_success(mock_bundles):
    supported_assets = specs.get_supported_assets(TEST_ITEM_TYPE)
    assert TEST_ASSET_TYPE in supported_assets


@respx.mock
def test_get_supported_assets_not_supported_item_type(mock_bundles):
    with pytest.raises(specs.SpecificationException):
        specs.get_supported_assets('notsupported')


@respx.mock
def test_validate_asset_type_supported(mock_bundles):
    """Ensures that a validated asset type for a given item type matches the
    the given asset type."""
    assert TEST_ASSET_TYPE == specs.validate_asset_type(
        TEST_ITEM_TYPE, TEST_ASSET_TYPE)


@respx.mock
def test_validate_asset_type_notsupported(mock_bundles):
    with pytest.raises(specs.SpecificationException):
        specs.validate_asset_type(TEST_ITEM_TYPE, 'notsupported')



"""Tests of the Planet Subscriptions API client."""

import csv
from datetime import datetime
from itertools import zip_longest
import json

from httpx import Response
import pytest
import respx
from respx.patterns import M

from planet.sync import Planet
from planet.clients.subscriptions import SubscriptionsClient
from planet.exceptions import APIError, PagingError, ServerError
from planet.http import Session

# M(url=TEST_URL) is case sensitive and matching a lower-case url, and it
# requires that there not be a '/'
TEST_URL = 'http://www.mocknotrealurl.com/api/path'

# Mock router representing an API that returns only a 500
# internal server error. The router is configured outside the test
# to help keep our test more readable.
failing_api_mock = respx.mock()
failing_api_mock.route(M(url__startswith=TEST_URL)).mock(
    return_value=Response(500, json={
        "code": 0, "message": "string"
    }))


def subscription_pages(status=None, size=40):
    """Helper for creating fake subscriptions listing pages."""
    all_subs = [{'id': str(i), 'status': 'running'} for i in range(1, 101)]
    select_subs = (sub for sub in all_subs
                   if not status or sub['status'] in status)
    pages = []
    for group in zip_longest(*[iter(select_subs)] * size):
        subs = list(sub for sub in group if sub is not None)
        page = {'subscriptions': subs, '_links': {}}
        pm = datetime.now().isoformat()
        if len(subs) == size:
            page['_links']['next'] = f'{TEST_URL}?pm={pm}'
        pages.append(page)
    return pages


# By default, respx's mock router asserts that all configured routes
# are followed in a test. Each of our parameterized tests follows only
# one route, so if we want to configure routes once and reuse them, we
# must disable the default.
api_mock = respx.mock(assert_all_called=False)

# 1. Request for status: running. Response has three pages.
api_mock.route(M(url=TEST_URL),
               M(params__contains={'status': 'running'})).mock(side_effect=[
                   Response(200, json=page)
                   for page in subscription_pages(status={'running'}, size=40)
               ])

# 2. Request for status: failed. Response has a single empty page.
api_mock.route(M(url=TEST_URL), M(params__contains={'status': 'failed'})).mock(
    side_effect=[Response(200, json={'subscriptions': []})])

# 3. Request for status: preparing. Response has a single empty page.
api_mock.route(M(url=TEST_URL),
               M(params__contains={'status': 'preparing'})).mock(
                   side_effect=[Response(200, json={'subscriptions': []})])

# 4. source_type: catalog requested. Response is the same as for 1.
api_mock.route(
    M(url=TEST_URL),
    M(params__contains={'source_type': 'catalog'})).mock(side_effect=[
        Response(200, json=page)
        for page in subscription_pages(status={'running'}, size=40)
    ])

# 5. source_type: soil_water_content requested. Response has a single empty page.
api_mock.route(M(url=TEST_URL),
               M(params__contains={'source_type': 'soil_water_content'})).mock(
                   side_effect=[Response(200, json={'subscriptions': []})])

# 6. All other parameters are used. Response has 2 subscriptions.
# The response is unrealistic here, but we are just testing the query parameter handling.
api_mock.route(
    M(url=TEST_URL),
    M(
        params__contains={
            'name': 'test xyz',
            'name__contains': 'xyz',
            'created': '2018-02-12T00:00:00Z/..',
            'updated': '../2018-03-18T12:31:12Z',
            'start_time': '2018-01-01T00:00:00Z',
            'end_time': '2022-01-01T00:00:00Z/2024-01-01T00:00:00Z',
            'hosting': 'true',
            'sort_by': 'name DESC',
        })).mock(side_effect=[
            Response(200, json={'subscriptions': [{
                'id': 1
            }, {
                'id': 2
            }]})
        ])

# 7. No status or source_type requested. Response is the same as for 1.
api_mock.route(M(url=TEST_URL)).mock(side_effect=[
    Response(200, json=page) for page in subscription_pages(size=40)
])


# The "creation", "update", and "cancel" mock APIs return submitted
# data to the caller. They are used to test methods that rely on POST,
# PATCH, or PUT.
def modify_response(request):
    if request.content:
        return Response(200, json=json.loads(request.content))
    else:
        return Response(200)


create_mock = respx.mock()
create_mock.route(M(url=TEST_URL),
                  method='POST').mock(side_effect=modify_response)

update_mock = respx.mock()
update_mock.route(M(url=f'{TEST_URL}/test'),
                  method='PUT').mock(side_effect=modify_response)

patch_mock = respx.mock()
patch_mock.route(M(url=f'{TEST_URL}/test'),
                 method='PATCH').mock(side_effect=modify_response)

cancel_mock = respx.mock()
cancel_mock.route(M(url=f'{TEST_URL}/test/cancel'),
                  method='POST').mock(side_effect=modify_response)

# Mock the subscription description API endpoint.
get_mock = respx.mock()
get_mock.route(
    M(url=f'{TEST_URL}/test'),
    method='GET').mock(return_value=Response(200,
                                             json={
                                                 'id': '42',
                                                 'name': 'test',
                                                 'delivery': 'yes, please',
                                                 'source': 'test'
                                             }))

summary_mock = respx.mock()
summary_mock.route(
    M(url=f'{TEST_URL}/summary'),
    method='GET').mock(return_value=Response(200,
                                             json={
                                                 "subscriptions": {
                                                     "preparing": 0,
                                                     "pending": 0,
                                                     "running": 0,
                                                     "completed": 0,
                                                     "cancelled": 0,
                                                     "suspended": 0,
                                                     "failed": 0
                                                 }
                                             }))

sub_summary_mock = respx.mock()
sub_summary_mock.route(
    M(url=f'{TEST_URL}/test/summary'),
    method='GET').mock(return_value=Response(200,
                                             json={
                                                 "results": {
                                                     "created": 0,
                                                     "queued": 0,
                                                     "processing": 0,
                                                     "failed": 0,
                                                     "cancelled": 0,
                                                     "success": 0
                                                 },
                                                 "subscription": {
                                                     "status": "pending"
                                                 }
                                             }))


def result_pages(status=None, size=40):
    """Helper for creating fake result listing pages."""
    all_results = [{'id': str(i), 'status': 'created'} for i in range(1, 101)]
    select_results = (result for result in all_results
                      if not status or result['status'] in status)
    pages = []
    for group in zip_longest(*[iter(select_results)] * size):
        results = list(result for result in group if result is not None)
        page = {'results': results, '_links': {}}
        pm = datetime.now().isoformat()
        if len(results) == size:
            url = f'{TEST_URL}/42/results?pm={pm}'
            page['_links']['next'] = url
        pages.append(page)
    return pages


# By default, respx's mock router asserts that all configured routes
# are followed in a test. Each of our parameterized tests follows only
# one route, so if we want to configure routes once and reuse them, we
# must disable the default.
res_api_mock = respx.mock(assert_all_called=False)

# 1. CSV results
res_api_mock.route(
    M(url__startswith=TEST_URL), M(params__contains={'format': 'csv'})).mock(
        side_effect=[Response(200, text="id,status\n1234-abcd,SUCCESS\n")])

# 2. Request for status: created. Response has three pages.
res_api_mock.route(
    M(url__startswith=TEST_URL),
    M(params__contains={'status': 'created'})).mock(side_effect=[
        Response(200, json=page)
        for page in result_pages(status={'created'}, size=40)
    ])

# 3. Request for status: queued. Response has a single empty page.
res_api_mock.route(M(url__startswith=TEST_URL),
                   M(params__contains={'status': 'queued'})).mock(
                       side_effect=[Response(200, json={'results': []})])

# 4. No status requested. Response is the same as for 1.
res_api_mock.route(M(url__startswith=TEST_URL)).mock(
    side_effect=[Response(200, json=page) for page in result_pages(size=40)])


@pytest.mark.anyio
@failing_api_mock
async def test_list_subscriptions_failure():
    """ServerError is raised if there is an internal server error (500)."""
    with pytest.raises(ServerError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = [sub async for sub in client.list_subscriptions()]


@pytest.mark.parametrize("status, count", [({"running"}, 100), ({"failed"}, 0),
                                           (None, 100)])
@pytest.mark.anyio
@api_mock
async def test_list_subscriptions_success(
    status,
    count,
):
    """Account subscriptions iterator yields expected descriptions."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        assert len([
            sub async for sub in client.list_subscriptions(status=status)
        ]) == count


@pytest.mark.parametrize("status, count", [({"running"}, 100), ({"failed"}, 0),
                                           (None, 100)])
@api_mock
def test_list_subscriptions_success_sync(
    status,
    count,
):
    """Account subscriptions iterator yields expected descriptions."""
    client = Planet()
    client.subscriptions._client._base_url = TEST_URL
    assert len(list(
        client.subscriptions.list_subscriptions(status=status))) == count


@pytest.mark.parametrize("source_type, count",
                         [("catalog", 100), ("soil_water_content", 0),
                          (None, 100)])
@pytest.mark.anyio
@api_mock
async def test_list_subscriptions_source_type_success(
    source_type,
    count,
):
    """Account subscriptions iterator yields expected descriptions."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        assert len([
            sub
            async for sub in client.list_subscriptions(source_type=source_type)
        ]) == count


@pytest.mark.anyio
@api_mock
async def test_list_subscriptions_filtering_and_sorting():
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        assert len([
            sub async for sub in client.list_subscriptions(
                name='test xyz', name__contains='xyz',
                created='2018-02-12T00:00:00Z/..',
                updated='../2018-03-18T12:31:12Z',
                start_time='2018-01-01T00:00:00Z',
                end_time='2022-01-01T00:00:00Z/2024-01-01T00:00:00Z',
                hosting=True, sort_by='name DESC')
        ]) == 2


@pytest.mark.parametrize("page_size, count", [(50, 100), (100, 100)])
@pytest.mark.anyio
@api_mock
async def test_list_subscriptions_page_size_success(page_size, count):
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        assert len([
            sub async for sub in client.list_subscriptions(page_size=page_size)
        ]) == count


@pytest.mark.anyio
@failing_api_mock
async def test_create_subscription_failure():
    """APIError is raised if there is a server error."""
    with pytest.raises(ServerError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = await client.create_subscription({"lol": "wut"})


@pytest.mark.anyio
@create_mock
async def test_create_subscription_success():
    """Subscription is created, description has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        sub = await client.create_subscription({
            'name': 'test', 'delivery': 'yes, please', 'source': 'test'
        })
        assert sub['name'] == 'test'


@create_mock
def test_create_subscription_success_sync():
    """Subscription is created, description has the expected items."""

    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    sub = pl.subscriptions.create_subscription({
        'name': 'test', 'delivery': 'yes, please', 'source': 'test'
    })
    assert sub['name'] == 'test'


@pytest.mark.anyio
@create_mock
async def test_create_subscription_with_hosting_success():
    """Subscription is created, description has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        sub = await client.create_subscription({
            'name': 'test', 'source': 'test', 'hosting': 'yes, please'
        })
        assert sub['name'] == 'test'


@create_mock
def test_create_subscription_with_hosting_success_sync():
    """Subscription is created, description has the expected items."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    sub = pl.subscriptions.create_subscription({
        'name': 'test', 'source': 'test', 'hosting': 'yes, please'
    })
    assert sub['name'] == 'test'


@pytest.mark.anyio
@failing_api_mock
async def test_cancel_subscription_failure():
    """APIError is raised if there is a server error."""
    with pytest.raises(ServerError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = await client.cancel_subscription("lolwut")


@pytest.mark.anyio
@cancel_mock
async def test_cancel_subscription_success():
    """Subscription is canceled, description has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        _ = await client.cancel_subscription("test")


@cancel_mock
def test_cancel_subscription_success_sync():
    """Subscription is canceled, description has the expected items."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    _ = pl.subscriptions.cancel_subscription("test")


@pytest.mark.anyio
@failing_api_mock
async def test_update_subscription_failure():
    """APIError is raised if there is a server error."""
    with pytest.raises(ServerError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = await client.update_subscription("lolwut", {})


@pytest.mark.anyio
@update_mock
async def test_update_subscription_success():
    """Subscription is updated, description has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        sub = await client.update_subscription(
            "test", {
                "name": "test", "delivery": "no, thanks", "source": "test"
            })
        assert sub["delivery"] == "no, thanks"


@update_mock
def test_update_subscription_success_sync():
    """Subscription is updated, description has the expected items."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    sub = pl.subscriptions.update_subscription(
        "test", {
            "name": "test", "delivery": "no, thanks", "source": "test"
        })
    assert sub["delivery"] == "no, thanks"


@pytest.mark.anyio
@patch_mock
async def test_patch_subscription_success():
    """Subscription is patched, description has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        sub = await client.patch_subscription("test", {"name": "test patch"})
        assert sub["name"] == "test patch"


@patch_mock
def test_patch_subscription_success_sync():
    """Subscription is patched, description has the expected items."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    sub = pl.subscriptions.patch_subscription("test", {"name": "test patch"})
    assert sub["name"] == "test patch"


@pytest.mark.anyio
@failing_api_mock
async def test_get_subscription_failure():
    """APIError is raised if there is a server error."""
    with pytest.raises(ServerError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = await client.get_subscription("lolwut")


@pytest.mark.anyio
@get_mock
async def test_get_subscription_success(monkeypatch):
    """Subscription description fetched, has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        sub = await client.get_subscription("test")
        assert sub['delivery'] == "yes, please"


@get_mock
def test_get_subscription_success_sync(monkeypatch):
    """Subscription description fetched, has the expected items."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    sub = pl.subscriptions.get_subscription("test")
    assert sub['delivery'] == "yes, please"


@pytest.mark.anyio
@failing_api_mock
async def test_get_results_failure():
    """APIError is raised if there is a server error."""
    with pytest.raises(APIError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = [res async for res in client.get_results("lolwut")]


@pytest.mark.anyio
@res_api_mock
async def test_get_results_success():
    """Subscription description fetched, has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        results = [res async for res in client.get_results("42")]
        assert len(results) == 100


@res_api_mock
def test_get_results_success_sync():
    """Subscription description fetched, has the expected items."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    results = list(pl.subscriptions.get_results("42"))
    assert len(results) == 100


@pytest.mark.anyio
@res_api_mock
async def test_get_results_csv():
    """Subscription CSV fetched, has the expected items."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        results = [res async for res in client.get_results_csv("42")]
        rows = list(csv.reader(results))
        assert rows == [['id', 'status'], ['1234-abcd', 'SUCCESS']]


@res_api_mock
def test_get_results_csv_sync():
    """Subscription CSV fetched, has the expected items."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    results = list(pl.subscriptions.get_results_csv("42"))
    rows = list(csv.reader(results))
    assert rows == [['id', 'status'], ['1234-abcd', 'SUCCESS']]


paging_cycle_api_mock = respx.mock()

# Identical next links is a hangup we want to avoid.
paging_cycle_api_mock.route(M(url__startswith=TEST_URL)).mock(side_effect=[
    Response(200,
             json={
                 'subscriptions': [{
                     'id': '1'
                 }], '_links': {
                     "next": TEST_URL
                 }
             }),
    Response(200,
             json={
                 'subscriptions': [{
                     'id': '2'
                 }], '_links': {
                     "next": TEST_URL
                 }
             })
])


@pytest.mark.anyio
@paging_cycle_api_mock
async def test_list_subscriptions_cycle_break():
    """PagingError is raised if there is a paging cycle."""
    with pytest.raises(PagingError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = [sub async for sub in client.list_subscriptions()]


@pytest.mark.anyio
@failing_api_mock
async def test_get_summary_failure():
    """ServerError is raised if there is an internal server error (500)."""
    with pytest.raises(ServerError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = await client.get_summary()


@pytest.mark.anyio
@summary_mock
async def test_get_summary():
    """Summary fetched, has expected totals, async."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        summary = await client.get_summary()
        assert summary == {
            "subscriptions": {
                "preparing": 0,
                "pending": 0,
                "running": 0,
                "completed": 0,
                "cancelled": 0,
                "suspended": 0,
                "failed": 0
            }
        }


@summary_mock
def test_get_summary_sync():
    """Summary fetched, has expected totals, sync."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    summary = pl.subscriptions.get_summary()
    assert summary == {
        "subscriptions": {
            "preparing": 0,
            "pending": 0,
            "running": 0,
            "completed": 0,
            "cancelled": 0,
            "suspended": 0,
            "failed": 0
        }
    }


@pytest.mark.anyio
@failing_api_mock
async def test_get_sub_summary_failure():
    """ServerError is raised if there is an internal server error (500)."""
    with pytest.raises(ServerError):
        async with Session() as session:
            client = SubscriptionsClient(session, base_url=TEST_URL)
            _ = await client.get_subscription_summary("test")


@pytest.mark.anyio
@sub_summary_mock
async def test_get_sub_summary():
    """Subscription summary fetched, has expected totals, async."""
    async with Session() as session:
        client = SubscriptionsClient(session, base_url=TEST_URL)
        summary = await client.get_subscription_summary("test")
        assert summary == {
            "results": {
                "created": 0,
                "queued": 0,
                "processing": 0,
                "failed": 0,
                "cancelled": 0,
                "success": 0
            },
            "subscription": {
                "status": "pending"
            }
        }


@sub_summary_mock
def test_get_sub_summary_sync():
    """Subscription summary fetched, has expected totals, sync."""
    pl = Planet()
    pl.subscriptions._client._base_url = TEST_URL
    summary = pl.subscriptions.get_subscription_summary("test")
    assert summary == {
        "results": {
            "created": 0,
            "queued": 0,
            "processing": 0,
            "failed": 0,
            "cancelled": 0,
            "success": 0
        },
        "subscription": {
            "status": "pending"
        }
    }



# Copyright 2021 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from http import HTTPStatus
import logging

import httpx
import jwt
import pytest
import respx

from planet import exceptions
from planet.auth import AuthClient

TEST_URL = 'http://MockNotRealURL/api/path'
TEST_LOGIN_URL = f'{TEST_URL}/login'

LOGGER = logging.getLogger(__name__)


@respx.mock
def test_AuthClient_success():
    payload = {'api_key': 'iamakey'}
    resp = {'token': jwt.encode(payload, 'key')}
    mock_resp = httpx.Response(HTTPStatus.OK, json=resp)
    respx.post(TEST_LOGIN_URL).return_value = mock_resp

    cl = AuthClient(base_url=TEST_URL)
    auth_data = cl.login('email', 'password')

    assert auth_data == payload


@respx.mock
def test_AuthClient_invalid_email():
    resp = {
        "errors": {
            "email": ["Not a valid email address."]
        },
        "message": "error validating request against UserAuthenticationSchema",
        "status": 400,
        "success": False
    }
    mock_resp = httpx.Response(400, json=resp)
    respx.post(TEST_LOGIN_URL).return_value = mock_resp

    cl = AuthClient(base_url=TEST_URL)
    with pytest.raises(exceptions.APIError,
                       match='Not a valid email address.'):
        _ = cl.login('email', 'password')


@respx.mock
def test_AuthClient_invalid_password():
    resp = {
        "errors": None,
        "message": "Invalid email or password",
        "status": 401,
        "success": False
    }
    mock_resp = httpx.Response(401, json=resp)
    respx.post(TEST_LOGIN_URL).return_value = mock_resp

    cl = AuthClient(base_url=TEST_URL)
    with pytest.raises(exceptions.APIError,
                       match='Incorrect email or password.'):
        _ = cl.login('email', 'password')



"""Subscriptions CLI"""
from contextlib import asynccontextmanager
from typing import List, Optional

import click

from . import types
from .cmds import coro, translate_exceptions
from .io import echo_json
from .options import limit, pretty
from .session import CliSession
from planet.clients.subscriptions import SubscriptionsClient
from .. import subscription_request
from ..subscription_request import sentinel_hub
from ..specs import FetchBundlesSpecError, get_item_types, SpecificationException, validate_item_type
from .validators import check_geom


def check_item_types(ctx, param, item_types) -> Optional[List[dict]]:
    """Validates each item types provided by comparing them to all supported
    item types."""
    try:
        for item_type in item_types:
            validate_item_type(item_type)
        return item_types
    except SpecificationException as e:
        raise click.BadParameter(str(e))
    except FetchBundlesSpecError as e:
        raise click.ClickException(str(e))


@asynccontextmanager
async def subscriptions_client(ctx):
    async with CliSession() as sess:
        cl = SubscriptionsClient(sess, base_url=ctx.obj['BASE_URL'])
        yield cl


@click.group()  # type: ignore
@click.pass_context
@click.option('-u',
              '--base-url',
              default=None,
              help='Assign custom base Subscriptions API URL.')
def subscriptions(ctx, base_url):
    """Commands for interacting with the Subscriptions API"""
    ctx.obj['BASE_URL'] = base_url


# We want our command to be known as "list" on the command line but
# don't want to clobber Python's built-in "list". We'll define the
# command function as "list_subscriptions".
@subscriptions.command(name="list")  # type: ignore
@pretty
@click.option(
    '--created',
    help="""Filter subscriptions by creation time or interval. See documentation
    for examples.""")
@click.option(
    '--end-time',
    help="""Filter subscriptions by end time or interval. See documentation
    for examples.""")
@click.option(
    '--hosting',
    type=click.BOOL,
    help="""Filter subscriptions by presence of hosting block (e.g. SentinelHub
    hosting).""")
@click.option(
    '--name-contains',
    help="""only return subscriptions with a name that contains the provided
    string.""")
@click.option('--name', help="filter by name")
@click.option(
    '--source-type',
    help="""Filter subscriptions by source type. See documentation for all
    available types. Default is all.""")
@click.option(
    '--start-time',
    help="""Filter subscriptions by start time or interval. See documentation
    for examples.""")
@click.option(
    '--status',
    type=click.Choice([
        "running",
        "cancelled",
        "preparing",
        "pending",
        "completed",
        "suspended",
        "failed"
    ]),
    multiple=True,
    help="Select subscriptions in one or more states. Default is all.")
@click.option('--sort-by',
              help="""fields to sort subscriptions by. Multiple fields can be
    specified, separated by commas. The sort direction can be specified by
    appending ' ASC' or ' DESC' to the field name. The default sort
    direction is ascending. When multiple fields are specified, the sort
    order is applied in the order the fields are listed.

    Supported fields: [name, created, updated, start_time, end_time].

    Example: 'name ASC,created DESC'""")
@click.option('--updated',
              help="""Filter subscriptions by update time or interval. See
    documentation
    for examples.""")
@limit
@click.option('--page-size',
              type=click.INT,
              help='Number of subscriptions to return per page.')
@click.pass_context
@translate_exceptions
@coro
async def list_subscriptions_cmd(ctx,
                                 created,
                                 end_time,
                                 hosting,
                                 name_contains,
                                 name,
                                 source_type,
                                 start_time,
                                 status,
                                 sort_by,
                                 updated,
                                 limit,
                                 page_size,
                                 pretty):
    """Prints a sequence of JSON-encoded Subscription descriptions."""
    async with subscriptions_client(ctx) as client:
        list_subscriptions_kwargs = {
            'created': created,
            'end_time': end_time,
            'hosting': hosting,
            'name__contains': name_contains,
            'name': name,
            'source_type': source_type,
            'start_time': start_time,
            'status': status,
            'sort_by': sort_by,
            'updated': updated,
            'limit': limit,
        }
        if page_size is not None:
            list_subscriptions_kwargs['page_size'] = page_size

        async for sub in client.list_subscriptions(
                **list_subscriptions_kwargs):
            echo_json(sub, pretty)


@subscriptions.command(name="create")  # type: ignore
@click.argument("request", type=types.JSON())
@click.option(
    "--hosting",
    type=click.Choice([
        "sentinel_hub",
    ]),
    default=None,
    help='Hosting type. Currently, only "sentinel_hub" is supported.',
)
@click.option("--collection-id",
              default=None,
              help='Collection ID for Sentinel Hub hosting. '
              'If omitted, a new collection will be created.')
@click.option(
    '--create-configuration',
    is_flag=True,
    help='Automatically create a layer configuration for your collection. '
    'If omitted, no configuration will be created.')
@pretty
@click.pass_context
@translate_exceptions
@coro
async def create_subscription_cmd(ctx, request, pretty, **kwargs):
    """Create a subscription.

    Submits a subscription request for creation and prints the created
    subscription description, optionally pretty-printed.

    REQUEST is the full description of the subscription to be created. It must
    be JSON and can be specified a json string, filename, or '-' for stdin.

    Other flag options are hosting, collection_id, and create_configuration.
    The hosting flag specifies the hosting type, the collection_id flag specifies the
    collection ID for Sentinel Hub, and the create_configuration flag specifies
    whether or not to create a layer configuration for your collection. If the
    collection_id is omitted, a new collection will be created. If the
    create_configuration flag is omitted, no configuration will be created. The
    collection_id flag and create_configuration flag cannot be used together.
    """

    hosting = kwargs.get("hosting", None)
    collection_id = kwargs.get("collection_id", None)
    create_configuration = kwargs.get('create_configuration', False)

    if hosting == "sentinel_hub":
        hosting_info = sentinel_hub(collection_id, create_configuration)
        request["hosting"] = hosting_info

    async with subscriptions_client(ctx) as client:
        sub = await client.create_subscription(request)
        echo_json(sub, pretty)


@subscriptions.command(name='cancel')  # type: ignore
@click.argument('subscription_id')
@pretty
@click.pass_context
@translate_exceptions
@coro
async def cancel_subscription_cmd(ctx, subscription_id, pretty):
    """Cancels a subscription and prints the API response."""
    async with subscriptions_client(ctx) as client:
        _ = await client.cancel_subscription(subscription_id)


@subscriptions.command(name='update')  # type: ignore
@click.argument('subscription_id')
@click.argument('request', type=types.JSON())
@pretty
@click.pass_context
@translate_exceptions
@coro
async def update_subscription_cmd(ctx, subscription_id, request, pretty):
    """Update a subscription via PUT.

    Updates a subscription and prints the updated subscription description,
    optionally pretty-printed.

    REQUEST is the full description of the updated subscription. It must be
    JSON and can be specified a json string, filename, or '-' for stdin.
    """
    async with subscriptions_client(ctx) as client:
        sub = await client.update_subscription(subscription_id, request)
        echo_json(sub, pretty)


@subscriptions.command(name='patch')  # type: ignore
@click.argument('subscription_id')
@click.argument('request', type=types.JSON())
@pretty
@click.pass_context
@translate_exceptions
@coro
async def patch_subscription_cmd(ctx, subscription_id, request, pretty):
    """Update a subscription via PATCH.

    Updates a subscription and prints the updated subscription description,
    optionally pretty-printed.

    REQUEST only requires the attributes to be changed. It must be
    JSON and can be specified a json string, filename, or '-' for stdin.
    """
    async with subscriptions_client(ctx) as client:
        sub = await client.patch_subscription(subscription_id, request)
        echo_json(sub, pretty)


@subscriptions.command(name='get')  # type: ignore
@click.argument('subscription_id')
@pretty
@click.pass_context
@translate_exceptions
@coro
async def get_subscription_cmd(ctx, subscription_id, pretty):
    """Get the description of a subscription."""
    async with subscriptions_client(ctx) as client:
        sub = await client.get_subscription(subscription_id)
        echo_json(sub, pretty)


@subscriptions.command(name='results')  # type: ignore
@click.argument('subscription_id')
@pretty
@click.option(
    '--status',
    type=click.Choice(["created", "queued", "processing", "failed",
                       "success"]),
    multiple=True,
    default=None,
    callback=(lambda ctx, param, value: set(value)),
    help="Select subscription results in one or more states. Default: all.")
@click.option('--csv',
              'csv_flag',
              is_flag=True,
              default=False,
              help="Get subscription results as comma-separated fields. When "
              "this flag is included, --limit is ignored")
@limit
# TODO: the following 3 options.
# –created: timestamp instant or range.
# –updated: timestamp instant or range.
# –completed: timestamp instant or range.
@click.pass_context
@translate_exceptions
@coro
async def list_subscription_results_cmd(ctx,
                                        subscription_id,
                                        pretty,
                                        status,
                                        csv_flag,
                                        limit):
    """Print the results of a subscription to stdout.

    The output of this command is a sequence of JSON objects (the
    default) or a sequence of comma-separated fields (when the --csv
    option is used), one result per line.

    Examples:

    \b
        planet subscriptions results SUBSCRIPTION_ID --status=success --limit 10

    Where SUBSCRIPTION_ID is the unique identifier for a subscription,
    this prints the last 10 successfully delivered results for that
    subscription as JSON objects.

    \b
        planet subscriptions results SUBSCRIPTION_ID --limit 0 --csv > results.csv

    Prints all results for a subscription and saves them to a CSV file.
    """
    async with subscriptions_client(ctx) as client:
        if csv_flag:
            async for result in client.get_results_csv(subscription_id,
                                                       status=status):
                click.echo(result)
        else:
            async for result in client.get_results(subscription_id,
                                                   status=status,
                                                   limit=limit):
                echo_json(result, pretty)


@subscriptions.command()  # type: ignore
@translate_exceptions
@click.option('--name',
              required=True,
              type=str,
              help='Subscription name. Does not need to be unique.')
@click.option('--source',
              required=True,
              type=types.JSON(),
              help='Source JSON. Can be a string, filename, or - for stdin.')
@click.option(
    '--delivery',
    type=types.JSON(),
    help=("Delivery configuration, including credentials for a cloud "
          "storage provider, to enable cloud delivery of data. Can be a "
          "JSON string, a filename, or '-' for stdin. "))
@click.option(
    '--notifications',
    type=types.JSON(),
    help='Notifications JSON. Can be a string, filename, or - for stdin.')
@click.option(
    '--tools',
    type=types.JSON(),
    help='Toolchain JSON. Can be a string, filename, or - for stdin.')
@click.option(
    '--hosting',
    default=None,
    type=click.Choice([
        "sentinel_hub",
    ]),
    help='Hosting configuration. Can be JSON, "sentinel_hub", or omitted.')
@click.option(
    '--clip-to-source',
    is_flag=True,
    default=False,
    help="Clip to the source geometry without specifying a clip tool.")
@click.option("--collection-id",
              default=None,
              help='Collection ID for Sentinel Hub hosting. '
              'If omitted, a new collection will be created.')
@click.option(
    '--create-configuration',
    is_flag=True,
    default=False,
    help='Automatically create a layer configuration for your collection. '
    'If omitted, no configuration will be created.')
@pretty
def request(name,
            source,
            delivery,
            notifications,
            tools,
            hosting,
            collection_id,
            create_configuration,
            clip_to_source,
            pretty):
    """Generate a subscriptions request.

    Note: the next version of the Subscription API will remove the clip
    tool option and always clip to the source geometry. Thus the
    --clip-to-source option is a preview of the next API version's
    default behavior.
    """

    res = subscription_request.build_request(
        name,
        source,
        delivery,
        notifications=notifications,
        tools=tools,
        hosting=hosting,
        collection_id=collection_id,
        create_configuration=create_configuration,
        clip_to_source=clip_to_source)
    echo_json(res, pretty)


@subscriptions.command()  # type: ignore
@translate_exceptions
@click.option('--item-types',
              required=True,
              help='Item type for requested item ids.',
              type=types.CommaSeparatedString(),
              callback=check_item_types)
@click.option('--asset-types',
              required=True,
              type=types.CommaSeparatedString(),
              help='One or more comma-separated asset types.')
@click.option(
    '--geometry',
    required=True,
    type=types.Geometry(),
    callback=check_geom,
    help="""Geometry of the area of interest of the subscription that will be
    used to determine matches. Can be a string, filename, or - for stdin.""")
@click.option('--start-time',
              required=True,
              type=types.DateTime(),
              help='Date and time to begin subscription.')
@click.option('--end-time',
              type=types.DateTime(),
              help='Date and time to end subscription.')
@click.option('--rrule',
              type=str,
              help='iCalendar recurrance rule to specify recurrances.')
@click.option(
    '--filter',
    type=types.JSON(),
    help='Search filter.  Can be a string, filename, or - for stdin.')
@click.option(
    '--publishing-stage',
    'publishing_stages',
    type=click.Choice(["preview", "standard", "finalized"]),
    multiple=True,
    help=("Subscribe to results at a particular publishing stage. Multiple "
          "instances of this option are allowed."))
@click.option('--time-range-type',
              type=click.Choice(["acquired", "published"]),
              help="Subscribe by acquisition time or time of publication.")
@pretty
def request_catalog(item_types,
                    asset_types,
                    geometry,
                    start_time,
                    end_time,
                    rrule,
                    filter,
                    publishing_stages,
                    time_range_type,
                    pretty):
    """Generate a subscriptions request catalog source description."""

    res = subscription_request.catalog_source(
        item_types,
        asset_types,
        geometry,
        start_time,
        end_time=end_time,
        rrule=rrule,
        filter=filter,
        publishing_stages=publishing_stages,
        time_range_type=time_range_type)
    echo_json(res, pretty)


@subscriptions.command()  # type: ignore
@translate_exceptions
@click.option(
    '--var-type',
    required=False,
    help='A Planetary Variable type. See documentation for all available types.'
)
@click.option(
    '--var-id',
    required=True,
    help='A Planetary Variable ID. See documentation for all available IDs.')
@click.option(
    '--geometry',
    required=True,
    type=types.Geometry(),
    callback=check_geom,
    help="""Geometry of the area of interest of the subscription that will be
    used to determine matches. Can be a string, filename, or - for stdin.""")
@click.option('--start-time',
              required=True,
              type=types.DateTime(),
              help='Date and time to begin subscription.')
@click.option('--end-time',
              type=types.DateTime(),
              help='Date and time to end subscription.')
@pretty
def request_pv(var_type, var_id, geometry, start_time, end_time, pretty):
    """Generate a Planetary Variable subscription source.

    Planetary Variables come in 4 types and are further subdivided
    within these types. See [Subscribing to Planetary Variables](https://developers.planet.com/docs/subscriptions/pvs-subs/#planetary-variables-types-and-ids)
    or the [OpenAPI spec](https://api.planet.com/subscriptions/v1/spec) for
    more details.
    """
    res = subscription_request.planetary_variable_source(
        var_type,
        var_id,
        geometry,
        start_time,
        end_time=end_time,
    )
    echo_json(res, pretty)


@subscriptions.command()  # type: ignore
@click.pass_context
@translate_exceptions
def item_types(ctx):
    """Show valid item types for catalog subscriptions."""
    click.echo("Valid item types:")
    for it in get_item_types():
        click.echo(f"- {it}")


@subscriptions.command()  # type: ignore
@click.option(
    '--subscription-id',
    default=None,
    help="""Optionally supply a subscription ID to summarize result counts
    by status. If omitted, the summary will be generated for all
    subscriptions the requester has created by status.""")
@pretty
@click.pass_context
@translate_exceptions
@coro
async def summarize(ctx, subscription_id, pretty):
    """Summarize the status of all subscriptions or the status of results for a single subscription"""
    async with subscriptions_client(ctx) as client:
        if subscription_id:
            summary = await client.get_subscription_summary(subscription_id)
        else:
            summary = await client.get_summary()
        echo_json(summary, pretty)



# Copyright 2022 Planet Labs, PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from contextlib import nullcontext as does_not_raise
from datetime import datetime
import json

import click
from click.testing import CliRunner
from click.exceptions import BadParameter
import pytest

from planet.cli import types


@pytest.mark.parametrize("input,expectation, expected",
                         [('a', does_not_raise(), ['a']),
                          ('a,b', does_not_raise(), ['a', 'b']),
                          ('a, b', does_not_raise(), ['a', 'b']),
                          ('a,,', pytest.raises(BadParameter), None),
                          ('', pytest.raises(BadParameter), None),
                          (['a'], does_not_raise(), ['a'])])  # yapf: disable
def test_cli_CommaSeparatedString(input, expectation, expected):
    with expectation:
        res = types.CommaSeparatedString().convert(input, None, None)

    if expected:
        assert res == expected


@pytest.mark.parametrize("input, expectation, expected",
                         [
                             ('1.0', does_not_raise(), [1.0]),
                             ('1,2.5', does_not_raise(), [1.0, 2.5]),
                             ('1, 2.5', does_not_raise(), [1.0, 2.5]),
                             ('foo, bar', pytest.raises(BadParameter), None),
                             ('1,,', pytest.raises(BadParameter), None),
                             ([1.0, 2.0], does_not_raise(), [1.0, 2.0]),
                         ])
def test_cli_CommaSeparatedFloat(input, expectation, expected):
    with expectation:
        res = types.CommaSeparatedFloat().convert(input, None, None)

    if expected:
        assert res == expected


@pytest.fixture(params=["stdin", "str", "file"])
def json_input_test_params(request, write_to_tmp_json_file):

    def _func(input_json):
        if request.param == 'file':
            filename = write_to_tmp_json_file(input_json, 'tmp.json')
            arg = str(filename)
            input = None
        elif request.param == 'str':
            arg = json.dumps(input_json)
            input = None
        elif request.param == 'stdin':
            arg = '-'
            input = json.dumps(input_json)

        return arg, input

    return _func


def test_cli_JSON_inputs(json_input_test_params):
    """Confirm that file, stdin, and str inputs are all supported"""

    @click.option('--foo', type=types.JSON())
    @click.command()
    def test(foo):
        click.echo(json.dumps(foo))

    test_json = {'a': 'b'}
    arg, input = json_input_test_params(test_json)
    result = CliRunner().invoke(test, args=[f'--foo={arg}'], input=input)
    assert result.exit_code == 0
    assert json.loads(result.output) == test_json


parametrize_json = pytest.mark.parametrize(
    "input, expectation, expected",
    [
        ('{"a":["b", "c"], "c":5}',
         does_not_raise(), {
             'a': ['b', 'c'], 'c': 5
         }),
        ('["b", {"c":5}]', does_not_raise(), ['b', {
            'c': 5
        }]),
        ('{"a":"b", foo:bar}', pytest.raises(BadParameter), None),
        ('{}', pytest.raises(BadParameter), None),
    ])


@parametrize_json
def test_cli_JSON_str(input, expectation, expected):
    with expectation:
        res = types.JSON().convert(input, None, None)

    if expected:
        assert res == expected


@parametrize_json
def test_cli_JSON_file_content(input, expectation, expected, tmp_path):
    filename = tmp_path / 'temp.json'
    with open(filename, 'w') as fp:
        fp.write(input)

    with expectation:
        res = types.JSON().convert(str(filename), None, None)

    if expected:
        assert res == expected


def test_cli_JSON_file_doesnotexist():
    with pytest.raises(BadParameter):
        types.JSON().convert('nonexistant.json', None, None)


@pytest.mark.parametrize("input, expectation",
                         [('gt', does_not_raise()), ('lt', does_not_raise()),
                          ('foobar', pytest.raises(BadParameter))])
def test_cli_Comparison(input, expectation):
    with expectation:
        res = types.Comparison().convert(input, None, None)
        assert res == input


@pytest.mark.parametrize("input, expectation",
                         [('gt', does_not_raise()),
                          ('lt', pytest.raises(BadParameter)),
                          ('foobar', pytest.raises(BadParameter))])
def test_cli_GTComparison(input, expectation):
    with expectation:
        res = types.GTComparison().convert(input, None, None)
        assert res == input


@pytest.mark.parametrize(
    "input, expectation, expected",
    [('2020-02-02', does_not_raise(), datetime(2020, 2, 2)),
     ('2021-02-03T01:40:07.359Z',
      does_not_raise(),
      datetime(2021, 2, 3, 1, 40, 7, 359000)),
     ('2022', pytest.raises(BadParameter), None),
     (datetime(2020, 2, 2), does_not_raise(), datetime(2020, 2, 2))])
def test_cli_DateTime(input, expectation, expected):
    with expectation:
        res = types.DateTime().convert(input, None, None)

    if expected:
        assert res == expected



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for CLI I/O"""
import json

import click


def echo_json(obj: object, pretty: bool = False) -> None:
    """

    Parameters:
        obj: any object serializeable to JSON
        pretty: whether to reformat for easy visualization

    Returns:
        None
    """
    if pretty:
        json_str = json.dumps(obj, indent=2, sort_keys=True)
        click.echo(json_str)
    else:
        click.echo(json.dumps(obj))



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for interacting with the orders api"""
from typing import Any, Callable, Dict, Iterator, List, Optional

from pathlib import Path
from ..http import Session
from planet.clients import OrdersClient


class OrdersAPI:
    """Orders API client"""

    _client: OrdersClient

    def __init__(self, session: Session, base_url: Optional[str] = None):
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to production orders API
                base url.
        """

        self._client = OrdersClient(session, base_url)

    def create_order(self, request: Dict) -> Dict:
        """Create an order.

        Example:

        ```python

        from planet import Planet, order_request

        def main():
             pl = Planet()
             image_ids = ["20200925_161029_69_2223"]
             request = order_request.build_request(
                 'test_order',
                 [order_request.product(image_ids, 'analytic_udm2', 'psscene')]
             )
             order = pl.orders.create_order(request)
        ```

        Parameters:
            request: order request definition (recommended to use the order_request module to build a request)

        Returns:
            JSON description of the created order

        Raises:
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(self._client.create_order(request))

    def get_order(self, order_id: str) -> Dict:
        """Get order details by Order ID.

        Parameters:
            order_id: The ID of the order

        Returns:
            JSON description of the order

        Raises:
            planet.exceptions.ClientError: If order_id is not a valid UUID.
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(self._client.get_order(order_id))

    def cancel_order(self, order_id: str) -> Dict[str, Any]:
        """Cancel a queued order.

        Parameters:
            order_id: The ID of the order

        Returns:
            Results of the cancel request

        Raises:
            planet.exceptions.ClientError: If order_id is not a valid UUID.
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(self._client.cancel_order(order_id))

    def cancel_orders(self,
                      order_ids: Optional[List[str]] = None) -> Dict[str, Any]:
        """Cancel queued orders in bulk.

        Parameters:
            order_ids: The IDs of the orders. If empty or None, all orders in a
                pre-running state will be cancelled.

        Returns:
            Results of the bulk cancel request

        Raises:
            planet.exceptions.ClientError: If an entry in order_ids is not a
                valid UUID.
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(self._client.cancel_orders(order_ids))

    def aggregated_order_stats(self) -> Dict[str, Any]:
        """Get aggregated counts of active orders.

        Returns:
            Aggregated order counts

        Raises:
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(self._client.aggregated_order_stats())

    def download_asset(self,
                       location: str,
                       filename: Optional[str] = None,
                       directory: Path = Path('.'),
                       overwrite: bool = False,
                       progress_bar: bool = True) -> Path:
        """Download ordered asset.

        Parameters:
            location: Download location url including download token.
            filename: Custom name to assign to downloaded file.
            directory: Base directory for file download. This directory will be
                created if it does not already exist.
            overwrite: Overwrite any existing files.
            progress_bar: Show progress bar during download.

        Returns:
            Path to downloaded file.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(
            self._client.download_asset(location,
                                        filename,
                                        directory,
                                        overwrite,
                                        progress_bar))

    def download_order(self,
                       order_id: str,
                       directory: Path = Path('.'),
                       overwrite: bool = False,
                       progress_bar: bool = False) -> List[Path]:
        """Download all assets in an order.

        Parameters:
            order_id: The ID of the order.
            directory: Base directory for file download. This directory must
                already exist.
            overwrite: Overwrite files if they already exist.
            progress_bar: Show progress bar during download.

        Returns:
            Paths to downloaded files.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If the order is not in a final
                state.
        """
        return self._client._call_sync(
            self._client.download_order(order_id,
                                        directory,
                                        overwrite,
                                        progress_bar))

    def validate_checksum(self, directory: Path, checksum: str):
        """Validate checksums of downloaded files against order manifest.

        For each file entry in the order manifest, the specified checksum given
        in the manifest file will be validated against the checksum calculated
        from the downloaded file.

        Parameters:
            directory: Path to order directory.
            checksum: The type of checksum hash- 'MD5' or 'SHA256'.

        Raises:
            planet.exceptions.ClientError: If a file is missing or if checksums
                do not match.
        """
        return self._client.validate_checksum(directory, checksum)

    def wait(self,
             order_id: str,
             state: Optional[str] = None,
             delay: int = 5,
             max_attempts: int = 200,
             callback: Optional[Callable[[str], None]] = None) -> str:
        """Wait until order reaches desired state.

        Returns the state of the order on the last poll.

        This function polls the Orders API to determine the order state, with
        the specified delay between each polling attempt, until the
        order reaches a final state, or earlier state, if specified.
        If the maximum number of attempts is reached before polling is
        complete, an exception is raised. Setting 'max_attempts' to zero will
        result in no limit on the number of attempts.

        Setting 'delay' to zero results in no delay between polling attempts.
        This will likely result in throttling by the Orders API, which has
        a rate limit of 10 requests per second. If many orders are being
        polled asynchronously, consider increasing the delay to avoid
        throttling.

        By default, polling completes when the order reaches a final state.
        If 'state' is given, polling will complete when the specified earlier
        state is reached or passed.

        Example:
            ```python
            from planet.reporting import StateBar

            with StateBar() as bar:
                wait(order_id, callback=bar.update_state)
            ```

        Parameters:
            order_id: The ID of the order.
            state: State prior to a final state that will end polling.
            delay: Time (in seconds) between polls.
            max_attempts: Maximum number of polls. Set to zero for no limit.
            callback: Function that handles state progress updates.

        Returns
            State of the order.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If order_id or state is not valid or
                if the maximum number of attempts is reached before the
                specified state or a final state is reached.
        """
        return self._client._call_sync(
            self._client.wait(order_id, state, delay, max_attempts, callback))

    def list_orders(self,
                    state: Optional[str] = None,
                    limit: int = 100,
                    source_type: Optional[str] = None,
                    name: Optional[str] = None,
                    name__contains: Optional[str] = None,
                    created_on: Optional[str] = None,
                    last_modified: Optional[str] = None,
                    hosting: Optional[bool] = None,
                    sort_by: Optional[str] = None) -> Iterator[dict]:
        """Iterate over the list of stored orders.

        By default, order descriptions are sorted by creation date with the last created
        order returned first.

        Note:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            state (str): filter by state.
            source_type (str): filter by source type.
            name (str): filter by name.
            name__contains (str): only include orders with names containing this string.
            created_on (str): filter by creation date-time or interval.
            last_modified (str): filter by last modified date-time or interval.
            hosting (bool): only return orders that contain a hosting block
                (e.g. SentinelHub hosting).
            sort_by (str): fields to sort orders by. Multiple fields can be specified,
                separated by commas. The sort direction can be specified by appending
                ' ASC' or ' DESC' to the field name. The default sort direction is
                ascending. When multiple fields are specified, the sort order is applied
                in the order the fields are listed.

                Supported fields: name, created_on, state, last_modified

                Examples:
                 * "name"
                 * "name DESC"
                 * "name,state DESC,last_modified"
            limit (int): maximum number of results to return. When set to 0, no
                maximum is applied.

        Datetime args (created_on and last_modified) can either be a date-time or an
        interval, open or closed. Date and time expressions adhere to RFC 3339. Open
        intervals are expressed using double-dots.

        Yields:
            Description of an order.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If state is not valid.
        """
        results = self._client.list_orders(state,
                                           limit,
                                           source_type,
                                           name,
                                           name__contains,
                                           created_on,
                                           last_modified,
                                           hosting,
                                           sort_by)

        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass



from pathlib import Path
import shutil

import nox

nox.options.stop_on_first_error = True
nox.options.reuse_existing_virtualenvs = False

nox.options.sessions = ['lint', 'analyze', 'test', 'coverage', 'docs']

source_files = ("planet", "examples", "tests", "setup.py", "noxfile.py")

BUILD_DIRS = ['build', 'dist']


@nox.session
def analyze(session):
    session.install(".[lint]")

    session.run("mypy", "--ignore-missing", "planet")


@nox.session
def coverage(session):
    session.install("-e", ".[test]")

    session.run('coverage',
                'run',
                '-m',
                'pytest',
                '-qq',
                '--no-header',
                '--no-summary',
                '--no-cov',
                '--ignore',
                'examples/')
    session.run('coverage', 'report')


@nox.session(python=["3.9", "3.10", "3.11", "3.12", "3.13"])
def test(session):
    session.run('python', '-m', 'ensurepip', '--upgrade')
    session.install('-U', 'setuptools')
    session.install(".[test]")

    options = session.posargs
    # -W=error raises pytest warnings to errors so they are caught by CI
    # to exclude some warnings, see
    # https://docs.python.org/3/library/warnings.html#temporarily-suppressing-warnings
    session.run('python',
                '-m',
                'pytest',
                '--ignore',
                'examples/',
                '-v',
                '-Werror',
                '-Wignore::DeprecationWarning:tqdm.std',
                *options)


@nox.session
def lint(session):
    session.install("-e", ".[lint]")

    session.run("flake8", *source_files)
    session.run('yapf', '--diff', '-r', *source_files)


@nox.session
def docs_test(session):
    session.install("-e", ".[docs]")

    options = session.posargs

    # Because these doc examples can be long-running, output
    # the INFO and above log messages so we know what's happening
    session.run('pytest',
                '--doctest-glob',
                '*.md',
                '--no-cov',
                '--ignore',
                'examples/',
                '--ignore',
                'tests/',
                '--log-cli-level=INFO',
                *options)


@nox.session
def docs(session):
    session.install("-e", ".[docs]")

    session.run("mkdocs", "build")


@nox.session
def watch(session):
    """Build and serve live docs for editing"""
    session.install("-e", ".[docs]")

    session.run("mkdocs", "serve")


@nox.session
def examples(session):
    session.install("-e", ".[test]")

    options = session.posargs

    # Because these example scripts can be long-running, output the
    # example's stdout so we know what's happening
    session.run('pytest', '--no-cov', 'examples/', '-s', *options)


@nox.session
def build(session):
    """Build package"""
    # check preexisting
    exist_but_should_not = [p for p in BUILD_DIRS if Path(p).is_dir()]
    if exist_but_should_not:
        session.error(f"Pre-existing {', '.join(exist_but_should_not)}. "
                      "Run clean session and try again")

    session.install('build', 'twine', 'check-wheel-contents')

    session.run(*'python -m build --sdist --wheel'.split())
    session.run('check-wheel-contents', 'dist')


@nox.session
def clean(session):
    """Remove build directories"""
    to_remove = [Path(d) for d in BUILD_DIRS if Path(d).is_dir()]
    for p in to_remove:
        shutil.rmtree(p)


@nox.session
def publish_testpypi(session):
    """Publish to TestPyPi using API token"""
    _publish(session, 'testpypi')


@nox.session
def publish_pypi(session):
    """Publish to PyPi using API token"""
    _publish(session, 'pypi')


def _publish(session, repository):
    missing = [p for p in BUILD_DIRS if not Path(p).is_dir()]
    if missing:
        session.error(
            f"Missing one or more build directories: {', '.join(missing)}. "
            "Run build session and try again")

    session.install('twine')

    files = [str(f) for f in Path('dist').iterdir()]
    session.run("twine", "check", *files)
    session.run("twine",
                "upload",
                f"--repository={repository}",
                '-u=__token__',
                *files)



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

import pytest

import planet


@pytest.fixture(autouse=True, scope='function')
def test_disable_limiter(monkeypatch):
    """Disable rate/max worker limiting on Session requests to avoid
    slowing down tests"""
    monkeypatch.setattr(planet.http, 'RATE_LIMIT', 0)
    monkeypatch.setattr(planet.http, 'MAX_ACTIVE', 0)


@pytest.fixture
async def session():
    async with planet.Session() as ps:
        yield ps


@pytest.fixture
def order_descriptions(order_description):
    order1 = order_description
    order1['id'] = 'oid1'
    order2 = copy.deepcopy(order_description)
    order2['id'] = 'oid2'
    order3 = copy.deepcopy(order_description)
    order3['id'] = 'oid3'
    return [order1, order2, order3]


@pytest.fixture
def search_results(search_result):
    result1 = search_result
    result1['id'] = 'id1'
    result2 = copy.deepcopy(search_result)
    result2['id'] = 'id2'
    result3 = copy.deepcopy(search_result)
    result3['id'] = 'id3'
    return [result1, result2, result3]



# Copyright 2015 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class PlanetError(Exception):
    """Root for all exceptions thrown by the SDK"""
    pass


class APIError(PlanetError):
    """General unexpected API response"""


class BadQuery(APIError):
    """Invalid inputs, HTTP 400"""
    pass


class InvalidAPIKey(APIError):
    """Invalid key, HTTP 401"""
    pass


class NoPermission(APIError):
    """Insufficient permissions, HTTP 403"""
    pass


class MissingResource(APIError):
    """Request for non existing resource, HTTP 404"""
    pass


class Conflict(APIError):
    """Request conflict with current state of the target resource, HTTP 409"""
    pass


class TooManyRequests(APIError):
    """Too many requests, HTTP 429"""
    pass


class OverQuota(APIError):
    """Quota exceeded, HTTP 429"""
    pass


class ServerError(APIError):
    """Unexpected internal server error, HTTP 500"""
    pass


class BadGateway(APIError):
    """Bad gateway, HTTP 502"""
    pass


class InvalidIdentity(APIError):
    """Raised when logging in with invalid credentials"""
    pass


class ClientError(PlanetError):
    """Exceptions thrown client-side"""
    pass


class AuthException(ClientError):
    """Exceptions encountered during authentication"""
    pass


class PagingError(ClientError):
    """For errors that occur during paging."""
    pass


class GeoJSONError(ClientError):
    """Errors that occur due to invalid GeoJSON"""


class FeatureError(ClientError):
    """Errors that occur due to incorrectly formatted feature reference"""



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality for validating against the Planet API specification."""
import httpx
import logging
import itertools

SUPPORTED_TOOLS = [
    'bandmath',
    'clip',
    'composite',
    'coregister',
    'file_format',
    'reproject',
    'tile',
    'toar',
    'harmonize'
]
SUPPORTED_ORDER_TYPES = ['full', 'partial']
SUPPORTED_ARCHIVE_TYPES = ['zip']
SUPPORTED_FILE_FORMATS = ['COG', 'PL_NITF']
HARMONIZE_TOOL_TARGET_SENSORS = ['Sentinel-2']
BAND_MATH_PIXEL_TYPE = ('Auto', '8U', '16U', '16S', '32R')
BAND_MATH_PIXEL_TYPE_DEFAULT = 'Auto'
PRODUCT_BUNDLES = None

LOGGER = logging.getLogger(__name__)


class FetchBundlesSpecError(Exception):
    """Custom exception for errors fetching the product bundles spec."""
    pass


class _LazyBundlesLoader:
    """Lazy load the product bundles spec from the API."""

    def __getitem__(self, key):
        bundles_spec_url = "https://api.planet.com/compute/ops/bundles/spec"
        retries = 2
        cache = getattr(self, "cache", None)
        if cache is None:
            for attempt in range(1, retries + 1):
                try:
                    response = httpx.get(bundles_spec_url)
                    response.raise_for_status()
                    break
                except:  # noqa: E722
                    if attempt == retries:
                        raise FetchBundlesSpecError(
                            "Unable to fetch spec from API to generate valid item types and bundles. Please retry!"
                        ) from None
            bundles = response.json()['bundles']
            item_types = set(
                itertools.chain.from_iterable(bundles[bundle]['assets'].keys()
                                              for bundle in bundles.keys()))
            cache = {
                'bundles': bundles,
                'bundle_names': bundles.keys(),
                'item_types': item_types
            }
            setattr(self, "cache", cache)
        return cache[key]


PRODUCT_BUNDLES = _LazyBundlesLoader()


class NoMatchException(Exception):
    """No match was found"""
    pass


class SpecificationException(Exception):
    """No match was found"""

    def __init__(self, value, supported, field_name):
        self.value = value
        self.supported = supported
        self.field_name = field_name
        self.opts = ', '.join(["'" + s + "'" for s in supported])

    def __str__(self):
        return (f'{self.field_name} - \'{self.value}\' is not one of '
                f'{self.opts}.')


def validate_bundle(item_type, bundle):
    validate_supported_bundles(item_type, bundle)
    return _validate_field(bundle,
                           PRODUCT_BUNDLES["bundle_names"],
                           'product_bundle')


def validate_item_type(item_type):
    return _validate_field(item_type,
                           PRODUCT_BUNDLES["item_types"],
                           'item_type')


def validate_data_item_type(item_type):
    """Validate and correct capitalization of data api item type."""
    return _validate_field(item_type, get_data_item_types(), 'item_type')


def get_data_item_types():
    """Item types supported by the data api."""
    # This is a quick-fix for gh-956, to be superseded by gh-960
    return PRODUCT_BUNDLES["item_types"] | {'SkySatVideo'}


def get_bundle_names():
    """Get all product bundle names."""
    return PRODUCT_BUNDLES["bundle_names"]


def validate_order_type(order_type):
    return _validate_field(order_type, SUPPORTED_ORDER_TYPES, 'order_type')


def validate_archive_type(archive_type):
    return _validate_field(archive_type,
                           SUPPORTED_ARCHIVE_TYPES,
                           'archive_type')


def validate_tool(tool):
    return _validate_field(tool, SUPPORTED_TOOLS, 'tool')


def validate_file_format(file_format):
    return _validate_field(file_format, SUPPORTED_FILE_FORMATS, 'file_format')


def _validate_field(value, supported, field_name):
    try:
        value = get_match(value, supported, field_name)
    except (NoMatchException):
        raise SpecificationException(value, supported, field_name)
    except FetchBundlesSpecError as e:
        raise e
    return value


def validate_supported_bundles(item_type, bundle):
    """Validate the provided item type and bundle combination are supported"""
    # get all the supported bundles for the given item type
    supported_bundles = []
    for product_bundle in PRODUCT_BUNDLES["bundle_names"]:
        available_item_types = set(
            PRODUCT_BUNDLES["bundles"][product_bundle]['assets'].keys())
        if item_type.lower() in [x.lower() for x in available_item_types]:
            supported_bundles.append(product_bundle)
    # validate the provided bundle is in the list of supported bundles
    return _validate_field(bundle, supported_bundles, 'bundle')


def validate_asset_type(item_type, asset_type):
    """Validates an asset type for a given item type."""
    # item type validation occurs inside get_supported_assets
    supported_assets = get_supported_assets(item_type)
    return _validate_field(asset_type, supported_assets, 'asset_type')


def get_match(test_entry, spec_entries, field_name):
    """Find and return matching spec entry regardless of capitalization.

    This is helpful for working with the API spec, where the capitalization
    is hard to remember but must be exact otherwise the API throws an
    exception."""
    try:
        match = next(e for e in spec_entries
                     if e.lower() == test_entry.lower())
    except (StopIteration):
        raise SpecificationException(test_entry, spec_entries, field_name)

    return match


def get_product_bundles(item_type=None):
    """Get product bundles supported by Orders API."""
    if item_type:
        supported_bundles = []
        for product_bundle in PRODUCT_BUNDLES["bundle_names"]:
            available_item_types = set(
                PRODUCT_BUNDLES["bundles"][product_bundle]['assets'].keys())
            if item_type.lower() in [x.lower() for x in available_item_types]:
                supported_bundles.append(product_bundle)
        return supported_bundles

    return PRODUCT_BUNDLES["bundle_names"]


def get_item_types(product_bundle=None):
    """If given product bundle, get specific item types supported by Orders
    API. Otherwise, get all item types supported by Orders API."""
    if product_bundle:
        return set(PRODUCT_BUNDLES["bundles"][product_bundle]['assets'].keys())

    return PRODUCT_BUNDLES["item_types"]


def get_supported_assets(item_type):
    """Get all assets supported by a given item type."""
    item_type = validate_item_type(item_type)
    supported_bundles = get_product_bundles(item_type)
    supported_assets = [
        PRODUCT_BUNDLES["bundles"][bundle]["assets"][item_type]
        for bundle in supported_bundles
    ]
    supported_assets = list(set(list(itertools.chain(*supported_assets))))

    return supported_assets
